"""Tests for QueryableRunStore functionality.

Tests both InMemoryRunStore and JsonFileRunStore implementations.
"""

import tempfile
from datetime import datetime, timezone, timedelta
from pathlib import Path

import pytest

from abstractruntime import RunState, RunStatus, WaitReason
from abstractruntime.core.models import WaitState
from abstractruntime.storage.in_memory import InMemoryRunStore
from abstractruntime.storage.json_files import JsonFileRunStore
from abstractruntime.storage.sqlite import SqliteDatabase, SqliteRunStore
from abstractruntime.storage.base import QueryableRunIndexStore, QueryableRunStore


def utc_now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def make_run(
    *,
    workflow_id: str = "test_wf",
    status: RunStatus = RunStatus.RUNNING,
    wait_reason: WaitReason | None = None,
    wait_until: str | None = None,
    updated_at: str | None = None,
) -> RunState:
    """Helper to create test RunState objects."""
    run = RunState.new(workflow_id=workflow_id, entry_node="start")
    run.status = status
    if updated_at:
        run.updated_at = updated_at

    if wait_reason is not None:
        run.waiting = WaitState(
            reason=wait_reason,
            wait_key=f"key_{run.run_id[:8]}",
            until=wait_until,
        )

    return run


class TestInMemoryRunStoreQuery:
    """Tests for InMemoryRunStore query methods."""

    def test_implements_queryable_protocol(self):
        """InMemoryRunStore should implement QueryableRunStore protocol."""
        store = InMemoryRunStore()
        assert isinstance(store, QueryableRunStore)

    def test_list_runs_empty(self):
        """list_runs on empty store returns empty list."""
        store = InMemoryRunStore()
        result = store.list_runs()
        assert result == []

    def test_list_runs_all(self):
        """list_runs without filters returns all runs."""
        store = InMemoryRunStore()

        run1 = make_run(status=RunStatus.RUNNING)
        run2 = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.USER)
        run3 = make_run(status=RunStatus.COMPLETED)

        store.save(run1)
        store.save(run2)
        store.save(run3)

        result = store.list_runs()
        assert len(result) == 3

    def test_list_runs_filter_by_status(self):
        """list_runs filters by status correctly."""
        store = InMemoryRunStore()

        run1 = make_run(status=RunStatus.RUNNING)
        run2 = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.USER)
        run3 = make_run(status=RunStatus.COMPLETED)

        store.save(run1)
        store.save(run2)
        store.save(run3)

        waiting = store.list_runs(status=RunStatus.WAITING)
        assert len(waiting) == 1
        assert waiting[0].run_id == run2.run_id

        completed = store.list_runs(status=RunStatus.COMPLETED)
        assert len(completed) == 1
        assert completed[0].run_id == run3.run_id

    def test_list_runs_filter_by_wait_reason(self):
        """list_runs filters by wait_reason correctly."""
        store = InMemoryRunStore()

        run1 = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.USER)
        run2 = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.UNTIL)
        run3 = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.EVENT)

        store.save(run1)
        store.save(run2)
        store.save(run3)

        user_waits = store.list_runs(wait_reason=WaitReason.USER)
        assert len(user_waits) == 1
        assert user_waits[0].run_id == run1.run_id

        until_waits = store.list_runs(wait_reason=WaitReason.UNTIL)
        assert len(until_waits) == 1
        assert until_waits[0].run_id == run2.run_id

    def test_list_runs_filter_by_workflow_id(self):
        """list_runs filters by workflow_id correctly."""
        store = InMemoryRunStore()

        run1 = make_run(workflow_id="wf_a")
        run2 = make_run(workflow_id="wf_b")
        run3 = make_run(workflow_id="wf_a")

        store.save(run1)
        store.save(run2)
        store.save(run3)

        wf_a_runs = store.list_runs(workflow_id="wf_a")
        assert len(wf_a_runs) == 2
        assert all(r.workflow_id == "wf_a" for r in wf_a_runs)

    def test_list_runs_combined_filters(self):
        """list_runs applies multiple filters correctly."""
        store = InMemoryRunStore()

        run1 = make_run(workflow_id="wf_a", status=RunStatus.WAITING, wait_reason=WaitReason.USER)
        run2 = make_run(workflow_id="wf_a", status=RunStatus.WAITING, wait_reason=WaitReason.UNTIL)
        run3 = make_run(workflow_id="wf_b", status=RunStatus.WAITING, wait_reason=WaitReason.USER)

        store.save(run1)
        store.save(run2)
        store.save(run3)

        result = store.list_runs(workflow_id="wf_a", wait_reason=WaitReason.USER)
        assert len(result) == 1
        assert result[0].run_id == run1.run_id

    def test_list_runs_respects_limit(self):
        """list_runs respects the limit parameter."""
        store = InMemoryRunStore()

        for _ in range(10):
            store.save(make_run())

        result = store.list_runs(limit=3)
        assert len(result) == 3

    def test_list_runs_ordered_by_updated_at_desc(self):
        """list_runs returns results ordered by updated_at descending."""
        store = InMemoryRunStore()

        run1 = make_run(updated_at="2025-01-01T00:00:00Z")
        run2 = make_run(updated_at="2025-01-03T00:00:00Z")
        run3 = make_run(updated_at="2025-01-02T00:00:00Z")

        store.save(run1)
        store.save(run2)
        store.save(run3)

        result = store.list_runs()
        assert result[0].run_id == run2.run_id  # Most recent first
        assert result[1].run_id == run3.run_id
        assert result[2].run_id == run1.run_id

    def test_list_due_wait_until_empty(self):
        """list_due_wait_until on empty store returns empty list."""
        store = InMemoryRunStore()
        result = store.list_due_wait_until(now_iso=utc_now_iso())
        assert result == []

    def test_list_due_wait_until_finds_due_runs(self):
        """list_due_wait_until finds runs whose wait time has passed."""
        store = InMemoryRunStore()

        past_time = "2020-01-01T00:00:00Z"
        future_time = "2099-01-01T00:00:00Z"

        run_due = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until=past_time,
        )
        run_not_due = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until=future_time,
        )
        run_wrong_reason = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.USER,
        )
        run_not_waiting = make_run(status=RunStatus.RUNNING)

        store.save(run_due)
        store.save(run_not_due)
        store.save(run_wrong_reason)
        store.save(run_not_waiting)

        result = store.list_due_wait_until(now_iso=utc_now_iso())
        assert len(result) == 1
        assert result[0].run_id == run_due.run_id

    def test_list_due_wait_until_ordered_by_until_asc(self):
        """list_due_wait_until returns results ordered by until ascending."""
        store = InMemoryRunStore()

        run1 = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until="2020-01-03T00:00:00Z",
        )
        run2 = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until="2020-01-01T00:00:00Z",
        )
        run3 = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until="2020-01-02T00:00:00Z",
        )

        store.save(run1)
        store.save(run2)
        store.save(run3)

        result = store.list_due_wait_until(now_iso=utc_now_iso())
        assert result[0].run_id == run2.run_id  # Earliest first
        assert result[1].run_id == run3.run_id
        assert result[2].run_id == run1.run_id


class TestJsonFileRunStoreQuery:
    """Tests for JsonFileRunStore query methods."""

    def test_implements_queryable_protocol(self, tmp_path: Path):
        """JsonFileRunStore should implement QueryableRunStore protocol."""
        store = JsonFileRunStore(tmp_path)
        assert isinstance(store, QueryableRunStore)

    def test_list_runs_empty(self, tmp_path: Path):
        """list_runs on empty store returns empty list."""
        store = JsonFileRunStore(tmp_path)
        result = store.list_runs()
        assert result == []

    def test_list_runs_filter_by_status(self, tmp_path: Path):
        """list_runs filters by status correctly."""
        store = JsonFileRunStore(tmp_path)

        run1 = make_run(status=RunStatus.RUNNING)
        run2 = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.USER)
        run3 = make_run(status=RunStatus.COMPLETED)

        store.save(run1)
        store.save(run2)
        store.save(run3)

        waiting = store.list_runs(status=RunStatus.WAITING)
        assert len(waiting) == 1
        assert waiting[0].run_id == run2.run_id

    def test_list_runs_filter_by_wait_reason(self, tmp_path: Path):
        """list_runs filters by wait_reason correctly."""
        store = JsonFileRunStore(tmp_path)

        run1 = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.USER)
        run2 = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.UNTIL)

        store.save(run1)
        store.save(run2)

        user_waits = store.list_runs(wait_reason=WaitReason.USER)
        assert len(user_waits) == 1
        assert user_waits[0].run_id == run1.run_id

    def test_list_runs_respects_limit(self, tmp_path: Path):
        """list_runs respects the limit parameter."""
        store = JsonFileRunStore(tmp_path)

        for _ in range(10):
            store.save(make_run())

        result = store.list_runs(limit=3)
        assert len(result) == 3

    def test_list_due_wait_until_finds_due_runs(self, tmp_path: Path):
        """list_due_wait_until finds runs whose wait time has passed."""
        store = JsonFileRunStore(tmp_path)

        past_time = "2020-01-01T00:00:00Z"
        future_time = "2099-01-01T00:00:00Z"

        run_due = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until=past_time,
        )
        run_not_due = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until=future_time,
        )

        store.save(run_due)
        store.save(run_not_due)

        result = store.list_due_wait_until(now_iso=utc_now_iso())
        assert len(result) == 1
        assert result[0].run_id == run_due.run_id

    def test_list_due_wait_until_ordered_by_until_asc(self, tmp_path: Path):
        """list_due_wait_until returns results ordered by until ascending."""
        store = JsonFileRunStore(tmp_path)

        run1 = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until="2020-01-03T00:00:00Z",
        )
        run2 = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until="2020-01-01T00:00:00Z",
        )

        store.save(run1)
        store.save(run2)

        result = store.list_due_wait_until(now_iso=utc_now_iso())
        assert result[0].run_id == run2.run_id  # Earliest first
        assert result[1].run_id == run1.run_id

    def test_persists_and_queries_across_instances(self, tmp_path: Path):
        """Queries work after creating a new store instance (persistence test)."""
        store1 = JsonFileRunStore(tmp_path)

        run = make_run(
            status=RunStatus.WAITING,
            wait_reason=WaitReason.UNTIL,
            wait_until="2020-01-01T00:00:00Z",
        )
        store1.save(run)

        # Create new instance pointing to same directory
        store2 = JsonFileRunStore(tmp_path)

        result = store2.list_due_wait_until(now_iso=utc_now_iso())
        assert len(result) == 1
        assert result[0].run_id == run.run_id


class TestSqliteRunStoreQuery:
    def test_implements_queryable_protocol(self, tmp_path: Path):
        store = SqliteRunStore(SqliteDatabase(tmp_path / "gateway.sqlite3"))
        assert isinstance(store, QueryableRunStore)
        assert isinstance(store, QueryableRunIndexStore)

    def test_list_run_index_filters_session_and_root_only(self, tmp_path: Path):
        store = SqliteRunStore(SqliteDatabase(tmp_path / "gateway.sqlite3"))

        root = make_run(status=RunStatus.COMPLETED, updated_at="2025-01-02T00:00:00Z")
        root.session_id = "sess_1"
        root.parent_run_id = None
        store.save(root)

        child = make_run(status=RunStatus.COMPLETED, updated_at="2025-01-03T00:00:00Z")
        child.session_id = "sess_1"
        child.parent_run_id = root.run_id
        store.save(child)

        other = make_run(status=RunStatus.COMPLETED, updated_at="2025-01-04T00:00:00Z")
        other.session_id = "sess_2"
        other.parent_run_id = None
        store.save(other)

        rows = store.list_run_index(session_id="sess_1", root_only=True, limit=10)
        assert [r.get("run_id") for r in rows] == [root.run_id]
        assert rows[0].get("parent_run_id") in {None, ""}

    def test_list_due_wait_until_uses_wait_index(self, tmp_path: Path):
        db_path = tmp_path / "gateway.sqlite3"
        store = SqliteRunStore(SqliteDatabase(db_path))

        past_time = "2020-01-01T00:00:00Z"
        run_due = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.UNTIL, wait_until=past_time)
        store.save(run_due)

        result = store.list_due_wait_until(now_iso=utc_now_iso())
        assert len(result) == 1
        assert result[0].run_id == run_due.run_id

        # When the run is no longer WAITING(UNTIL), it must be removed from the wait index.
        run_due.status = RunStatus.RUNNING
        run_due.waiting = None
        store.save(run_due)
        assert store.list_due_wait_until(now_iso=utc_now_iso()) == []

    def test_persists_across_instances(self, tmp_path: Path):
        db_path = tmp_path / "gateway.sqlite3"
        store1 = SqliteRunStore(SqliteDatabase(db_path))

        run = make_run(status=RunStatus.WAITING, wait_reason=WaitReason.UNTIL, wait_until="2020-01-01T00:00:00Z")
        store1.save(run)

        store2 = SqliteRunStore(SqliteDatabase(db_path))
        result = store2.list_due_wait_until(now_iso=utc_now_iso())
        assert len(result) == 1
        assert result[0].run_id == run.run_id


@pytest.mark.parametrize("store_kind", ["memory", "json", "sqlite"])
def test_run_index_includes_sanitized_lifecycle(store_kind: str, tmp_path: Path):
    if store_kind == "memory":
        store = InMemoryRunStore()
    elif store_kind == "json":
        store = JsonFileRunStore(tmp_path / "runs")
    else:
        store = SqliteRunStore(SqliteDatabase(tmp_path / "gateway.sqlite3"))

    run = make_run(status=RunStatus.COMPLETED)
    run.session_id = "sess-lifecycle"
    run.vars["_run_lifecycle"] = {
        "source": "abstractflow.editor",
        "purpose": "draft_test",
        "visibility": "private",
        "retention": {"mode": "ephemeral", "ttl_s": "120"},
        "ignored": {"private": True},
    }
    store.save(run)

    rows = store.list_run_index(session_id="sess-lifecycle", root_only=True, limit=10)
    assert len(rows) == 1
    assert rows[0]["run_lifecycle"] == {
        "source": "abstractflow.editor",
        "purpose": "draft_test",
        "visibility": "private",
        "retention": {"mode": "ephemeral", "ttl_s": 120},
    }



"""Runtime adapters around AbstractCore's public output-selector contract."""

from __future__ import annotations

from typing import Any, Dict, List, Optional

_MIN_ABSTRACTCORE_VERSION = (2, 13, 37)
_MIN_ABSTRACTCORE_VERSION_TEXT = "2.13.38"


def _version_tuple(value: str) -> tuple[int, int, int]:
    parts: list[int] = []
    for raw in str(value or "").split(".")[:3]:
        digits = []
        for char in raw:
            if not char.isdigit():
                break
            digits.append(char)
        parts.append(int("".join(digits) or "0"))
    while len(parts) < 3:
        parts.append(0)
    return parts[0], parts[1], parts[2]


try:
    from abstractcore.utils.version import __version__ as _abstractcore_version
    from abstractcore.core.output_specs import (
        is_output_request,
        normalize_output_spec,
        normalize_output_specs,
        output_has_generated_media,
        output_requires_non_chat_dispatch,
        strip_runtime_output_metadata,
    )
except ImportError as exc:  # pragma: no cover - import-time dependency guard
    raise ImportError(
        f"abstractruntime.integrations.abstractcore requires abstractcore>={_MIN_ABSTRACTCORE_VERSION_TEXT} "
        "for the public output selector contract."
    ) from exc

if _version_tuple(_abstractcore_version) < _MIN_ABSTRACTCORE_VERSION:  # pragma: no cover - environment guard
    raise ImportError(
        f"abstractruntime.integrations.abstractcore requires abstractcore>={_MIN_ABSTRACTCORE_VERSION_TEXT} "
        f"for the current server-auth, provider-key, generated-media, and capability-catalog contracts; "
        f"found abstractcore {_abstractcore_version}."
    )


def is_abstractcore_output_request(output: Any) -> bool:
    """Return True when `output` is AbstractCore's multimodal output selector."""

    return is_output_request(output)


def normalize_output_spec_for_runtime(output: Any) -> Dict[str, Any]:
    """Normalize an AbstractCore output selector using the same public aliases as core."""

    return normalize_output_spec(output)


def normalize_output_specs_for_runtime(output: Any) -> List[Dict[str, Any]]:
    return normalize_output_specs(output)


def strip_runtime_output_metadata_for_core(output: Any) -> Any:
    """Keep runtime artifact metadata out of AbstractCore capability kwargs."""

    return strip_runtime_output_metadata(output)


def output_runtime_metadata(output: Any) -> tuple[Optional[str], Dict[str, Any]]:
    """Extract runtime storage metadata from an AbstractCore output selector."""

    if not is_abstractcore_output_request(output):
        return None, {}

    specs = normalize_output_specs_for_runtime(output)
    run_id: Optional[str] = None
    tags: Dict[str, Any] = {}
    for spec in specs:
        raw_run_id = spec.get("run_id")
        if run_id is None and isinstance(raw_run_id, str) and raw_run_id.strip():
            run_id = raw_run_id.strip()

        raw_tags = spec.get("tags")
        if isinstance(raw_tags, dict):
            tags.update({str(k): str(v) for k, v in raw_tags.items() if k is not None and v is not None})

    return run_id, tags


def output_request_has_generated_media(output: Any) -> bool:
    """Return True when an output selector can produce generated binary media."""

    return output_has_generated_media(output)


def output_request_has_non_text_result(output: Any) -> bool:
    return output_requires_non_chat_dispatch(output)



"""WorkflowBundleRegistry (disk) for installed `.flow` bundles.

This registry is a host-side convenience layer:
- stores `.flow` bundles in a directory
- provides interface discovery from bundle manifests
- resolves bundle refs like `bundle_id@version` (or latest version)

Design goals:
- stdlib-only (to keep AbstractRuntime minimal)
- no global state; callers choose the directory
- correctness-first (scan bundles; optional caching can be added later)
"""

from __future__ import annotations

import hashlib
import os
import re
import shutil
import tempfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple

from .models import WorkflowBundleEntrypoint, WorkflowBundleManifest
from .reader import open_workflow_bundle


class WorkflowBundleRegistryError(ValueError):
    """Raised when registry operations fail (install/remove/resolve)."""


_BUNDLE_ID_SAFE_RE = re.compile(r"[^a-zA-Z0-9_-]+")
_BUNDLE_VERSION_SAFE_RE = re.compile(r"[^a-zA-Z0-9_.-]+")


def sanitize_bundle_id(raw: str) -> str:
    s = str(raw or "").strip()
    if not s:
        return ""
    s = _BUNDLE_ID_SAFE_RE.sub("-", s)
    s = re.sub(r"-{2,}", "-", s).strip("-")
    return s


def sanitize_bundle_version(raw: str) -> str:
    s = str(raw or "").strip()
    if not s:
        return ""
    s = _BUNDLE_VERSION_SAFE_RE.sub("-", s)
    s = re.sub(r"-{2,}", "-", s).strip("-")
    return s


def _try_parse_semver(v: str) -> Optional[Tuple[int, int, int]]:
    s = str(v or "").strip()
    if not s:
        return None
    parts = [p.strip() for p in s.split(".")]
    if not parts or any(not p for p in parts):
        return None
    nums: list[int] = []
    for p in parts:
        if not p.isdigit():
            return None
        nums.append(int(p))
    while len(nums) < 3:
        nums.append(0)
    return (nums[0], nums[1], nums[2])


def default_workflow_bundles_dir() -> Path:
    """Resolve the default bundles directory for `.flow` bundles.

    Priority:
    1) `ABSTRACTFRAMEWORK_WORKFLOWS_DIR` (shared/cross-package)
    2) `ABSTRACTFLOW_PUBLISH_DIR` (authoring host publish)
    3) `ABSTRACTFLOW_FLOWS_DIR` (legacy)
    4) repo/dev fallback: `./flows/bundles/` if it exists
    5) user default: `~/.abstractframework/workflows/`
    """
    env_candidates = (
        "ABSTRACTFRAMEWORK_WORKFLOWS_DIR",
        "ABSTRACTFLOW_PUBLISH_DIR",
        "ABSTRACTFLOW_FLOWS_DIR",
    )
    for name in env_candidates:
        v = os.getenv(name)
        if isinstance(v, str) and v.strip():
            return Path(v.strip()).expanduser().resolve()

    dev = Path("flows") / "bundles"
    if dev.exists() and dev.is_dir():
        return dev.resolve()

    return (Path.home() / ".abstractframework" / "workflows").resolve()


def _sha256_file(path: Path) -> str:
    h = hashlib.sha256()
    with path.open("rb") as f:
        for chunk in iter(lambda: f.read(1024 * 1024), b""):
            h.update(chunk)
    return h.hexdigest()


@dataclass(frozen=True)
class InstalledWorkflowBundle:
    bundle_id: str
    bundle_version: str
    path: Path
    manifest: WorkflowBundleManifest
    sha256: Optional[str] = None

    @property
    def bundle_ref(self) -> str:
        return f"{self.bundle_id}@{self.bundle_version}"


@dataclass(frozen=True)
class WorkflowEntrypointRef:
    bundle_id: str
    bundle_version: str
    flow_id: str
    name: str = ""
    description: str = ""
    interfaces: List[str] = field(default_factory=list)
    is_default: bool = False

    @property
    def bundle_ref(self) -> str:
        return f"{self.bundle_id}@{self.bundle_version}"

    @property
    def workflow_id(self) -> str:
        return f"{self.bundle_ref}:{self.flow_id}"


def _split_bundle_ref(raw: str) -> tuple[str, Optional[str]]:
    s = str(raw or "").strip()
    if not s:
        return ("", None)
    # Allow passing a bundle ref with a `.flow` suffix (e.g. "basic-agent.flow")
    # when referring to an installed bundle id (not a filesystem path).
    if s.lower().endswith(".flow") and "/" not in s and "\\" not in s:
        s = s[:-5]
    if "@" not in s:
        return (s, None)
    a, b = s.split("@", 1)
    a = a.strip()
    b = b.strip()
    if not a:
        return ("", None)
    if not b:
        return (a, None)
    return (a, b)


def _pick_latest_version(bundles_by_version: Dict[str, InstalledWorkflowBundle]) -> Optional[str]:
    items = [(str(ver), b) for ver, b in (bundles_by_version or {}).items() if isinstance(ver, str)]
    if not items:
        return None

    if all(_try_parse_semver(ver) is not None for ver, _ in items):
        return max(items, key=lambda x: _try_parse_semver(x[0]) or (0, 0, 0))[0]

    # Fallback: prefer newest created_at.
    def _key(x: tuple[str, InstalledWorkflowBundle]) -> tuple[str, str]:
        ver, b = x
        created = str(getattr(getattr(b, "manifest", None), "created_at", "") or "")
        return (created, ver)

    return max(items, key=_key)[0]


class WorkflowBundleRegistry:
    """A directory-backed registry for `.flow` bundles."""

    def __init__(self, bundles_dir: str | Path | None = None) -> None:
        base = Path(bundles_dir).expanduser() if bundles_dir is not None else default_workflow_bundles_dir()
        self.bundles_dir = base.expanduser().resolve()

    def ensure_dir(self) -> Path:
        self.bundles_dir.mkdir(parents=True, exist_ok=True)
        return self.bundles_dir

    def scan(self) -> List[InstalledWorkflowBundle]:
        """Scan the bundles directory for `.flow` bundles (best-effort)."""
        if not self.bundles_dir.exists() or not self.bundles_dir.is_dir():
            return []

        out: list[InstalledWorkflowBundle] = []
        for path in sorted(self.bundles_dir.glob("*.flow")):
            if not path.is_file():
                continue
            try:
                bundle = open_workflow_bundle(path)
            except Exception:
                continue
            man = bundle.manifest
            bid = str(getattr(man, "bundle_id", "") or "").strip()
            bver = str(getattr(man, "bundle_version", "") or "0.0.0").strip() or "0.0.0"
            if not bid:
                continue
            out.append(InstalledWorkflowBundle(bundle_id=bid, bundle_version=bver, path=path.resolve(), manifest=man))
        return out

    def bundles_by_id(self) -> Dict[str, Dict[str, InstalledWorkflowBundle]]:
        out: Dict[str, Dict[str, InstalledWorkflowBundle]] = {}
        for b in self.scan():
            out.setdefault(b.bundle_id, {})[b.bundle_version] = b
        return out

    def resolve_bundle(self, bundle_ref: str) -> InstalledWorkflowBundle:
        """Resolve `bundle_id[@version]` to an installed bundle (prefers latest)."""
        bid, ver = _split_bundle_ref(bundle_ref)
        if not bid:
            raise WorkflowBundleRegistryError("bundle_ref must be 'bundle_id' or 'bundle_id@version'")

        bundles = self.bundles_by_id().get(bid) or {}
        if not bundles:
            raise WorkflowBundleRegistryError(f"Bundle '{bid}' is not installed in {self.bundles_dir}")

        if ver:
            b = bundles.get(ver)
            if b is None:
                available = ", ".join(sorted(bundles.keys()))
                raise WorkflowBundleRegistryError(f"Bundle '{bid}@{ver}' not found (available: {available})")
            return b

        latest = _pick_latest_version(bundles)
        if not latest:
            raise WorkflowBundleRegistryError(f"Bundle '{bid}' has no versions installed")
        return bundles[latest]

    def resolve_entrypoint(
        self,
        ref: str,
        *,
        interface: Optional[str] = None,
    ) -> WorkflowEntrypointRef:
        """Resolve a workflow reference to a bundle entrypoint.

        Supported refs:
        - `bundle_id` (uses manifest.default_entrypoint if set, else first entrypoint)
        - `bundle_id@version`
        - `bundle_id[:flow_id]` or `bundle_id@version:flow_id`
        - `entrypoint_name` (unique match across bundles, best-effort)
        """
        s = str(ref or "").strip()
        if not s:
            raise WorkflowBundleRegistryError("workflow reference is required")

        bundle_flow_id: Optional[str] = None
        # Support bundle_id:flow_id (avoid clobbering Windows drive letters).
        if ":" in s and not re.match(r"^[A-Za-z]:[\\\\/]", s):
            left, right = s.split(":", 1)
            if left.strip() and right.strip():
                s = left.strip()
                bundle_flow_id = right.strip()

        # If the caller passed a `.flow` filename (not found relative to CWD),
        # try resolving it within the registry directory.
        if s.lower().endswith(".flow") and "/" not in s and "\\" not in s:
            candidate = (self.bundles_dir / s).expanduser()
            if candidate.exists() and candidate.is_file():
                try:
                    b = open_workflow_bundle(candidate)
                    man = b.manifest
                    return self._entrypoint_from_manifest(
                        manifest=man,
                        bundle_id=str(getattr(man, "bundle_id", "") or "").strip(),
                        bundle_version=str(getattr(man, "bundle_version", "") or "0.0.0").strip() or "0.0.0",
                        flow_id=bundle_flow_id,
                        interface=interface,
                    )
                except Exception:
                    pass

        # Fast path: resolve by bundle ref.
        try:
            bundle = self.resolve_bundle(s)
            man = bundle.manifest
            return self._entrypoint_from_manifest(
                manifest=man,
                bundle_id=bundle.bundle_id,
                bundle_version=bundle.bundle_version,
                flow_id=bundle_flow_id,
                interface=interface,
            )
        except WorkflowBundleRegistryError:
            pass

        # Best-effort: resolve by entrypoint name.
        needle = s.casefold()
        matches: list[WorkflowEntrypointRef] = []
        for b in self.scan():
            man = b.manifest
            for ep in list(getattr(man, "entrypoints", None) or []):
                ep_name = str(getattr(ep, "name", "") or "").strip()
                if not ep_name or ep_name.casefold() != needle:
                    continue
                epr = self._entrypoint_from_entrypoint(
                    ep=ep,
                    bundle_id=b.bundle_id,
                    bundle_version=b.bundle_version,
                    is_default=str(getattr(man, "default_entrypoint", "") or "").strip() == str(getattr(ep, "flow_id", "") or "").strip(),
                )
                if interface and interface not in epr.interfaces:
                    continue
                matches.append(epr)

        if not matches:
            raise WorkflowBundleRegistryError(f"Workflow '{ref}' not found in {self.bundles_dir}")

        if len(matches) > 1:
            options = ", ".join(sorted({m.bundle_ref for m in matches}))
            raise WorkflowBundleRegistryError(
                f"Workflow name '{ref}' matches multiple bundles ({options}); use bundle_id[@version] instead"
            )
        return matches[0]

    def list_entrypoints(
        self,
        *,
        interface: Optional[str] = None,
        latest_only: bool = True,
    ) -> List[WorkflowEntrypointRef]:
        """List available entrypoints (optionally filtered by interface)."""
        bundles = self.bundles_by_id()
        selected: Iterable[InstalledWorkflowBundle]
        if latest_only:
            latest: list[InstalledWorkflowBundle] = []
            for bid, versions in bundles.items():
                latest_ver = _pick_latest_version(versions)
                if latest_ver:
                    latest.append(versions[latest_ver])
            selected = latest
        else:
            selected = [b for versions in bundles.values() for b in versions.values()]

        out: list[WorkflowEntrypointRef] = []
        for b in selected:
            man = b.manifest
            default_fid = str(getattr(man, "default_entrypoint", "") or "").strip()
            for ep in list(getattr(man, "entrypoints", None) or []):
                epr = self._entrypoint_from_entrypoint(
                    ep=ep,
                    bundle_id=b.bundle_id,
                    bundle_version=b.bundle_version,
                    is_default=bool(default_fid and str(getattr(ep, "flow_id", "") or "").strip() == default_fid),
                )
                if interface and interface not in epr.interfaces:
                    continue
                out.append(epr)

        out.sort(key=lambda e: (e.bundle_id, e.bundle_version, e.name or e.flow_id))
        return out

    def install(self, source: str | Path, *, overwrite: bool = False) -> InstalledWorkflowBundle:
        """Install a `.flow` bundle into the registry directory."""
        src = Path(source).expanduser().resolve()
        if not src.exists() or not src.is_file():
            raise WorkflowBundleRegistryError(f"Bundle not found: {src}")

        bundle = open_workflow_bundle(src)
        man = bundle.manifest
        bid = str(getattr(man, "bundle_id", "") or "").strip()
        bver = str(getattr(man, "bundle_version", "") or "0.0.0").strip() or "0.0.0"
        if not bid:
            raise WorkflowBundleRegistryError(f"Bundle '{src}' has empty manifest.bundle_id")

        safe_id = sanitize_bundle_id(bid)
        safe_ver = sanitize_bundle_version(bver)
        if safe_id != bid:
            raise WorkflowBundleRegistryError(
                f"Bundle id '{bid}' contains unsafe characters. "
                f"Publish with a safe bundle_id (suggested: '{safe_id}')."
            )
        if safe_ver != bver:
            raise WorkflowBundleRegistryError(
                f"Bundle version '{bver}' contains unsafe characters. "
                f"Publish with a safe bundle_version (suggested: '{safe_ver}')."
            )

        self.ensure_dir()
        dest = (self.bundles_dir / f"{bid}@{bver}.flow").resolve()
        if dest.exists():
            if not overwrite:
                raise WorkflowBundleRegistryError(f"Bundle already installed: {dest}")
            try:
                dest.unlink()
            except Exception as e:
                raise WorkflowBundleRegistryError(f"Failed removing existing bundle: {dest} ({e})") from e

        try:
            shutil.copy2(src, dest)
        except Exception as e:
            raise WorkflowBundleRegistryError(f"Failed installing bundle to {dest}: {e}") from e

        try:
            bundle2 = open_workflow_bundle(dest)
            sha = _sha256_file(dest)
            return InstalledWorkflowBundle(
                bundle_id=str(bundle2.manifest.bundle_id),
                bundle_version=str(bundle2.manifest.bundle_version),
                path=dest,
                manifest=bundle2.manifest,
                sha256=sha,
            )
        except Exception:
            # Best-effort: return what we know even if re-open/hash fails.
            return InstalledWorkflowBundle(bundle_id=bid, bundle_version=bver, path=dest, manifest=man)

    def install_bytes(
        self,
        content: bytes,
        *,
        filename_hint: str = "upload.flow",
        overwrite: bool = False,
    ) -> InstalledWorkflowBundle:
        """Install bundle bytes into the registry directory.

        This is primarily intended for hosts that receive `.flow` bytes over the network.
        """
        data = bytes(content or b"")
        if not data:
            raise WorkflowBundleRegistryError("Bundle content is empty")

        self.ensure_dir()
        suffix = ".flow" if str(filename_hint or "").lower().endswith(".flow") else ".flow"
        try:
            fd, tmp_path_str = tempfile.mkstemp(prefix=".upload_", suffix=suffix, dir=str(self.bundles_dir))
            tmp_path = Path(tmp_path_str)
            with os.fdopen(fd, "wb") as f:
                f.write(data)
        except Exception as e:
            raise WorkflowBundleRegistryError(f"Failed writing upload temp file: {e}") from e

        try:
            bundle = open_workflow_bundle(tmp_path)
            man = bundle.manifest
            bid = str(getattr(man, "bundle_id", "") or "").strip()
            bver = str(getattr(man, "bundle_version", "") or "0.0.0").strip() or "0.0.0"
            if not bid:
                raise WorkflowBundleRegistryError("Uploaded bundle has empty manifest.bundle_id")

            safe_id = sanitize_bundle_id(bid)
            safe_ver = sanitize_bundle_version(bver)
            if safe_id != bid:
                raise WorkflowBundleRegistryError(
                    f"Bundle id '{bid}' contains unsafe characters. "
                    f"Publish with a safe bundle_id (suggested: '{safe_id}')."
                )
            if safe_ver != bver:
                raise WorkflowBundleRegistryError(
                    f"Bundle version '{bver}' contains unsafe characters. "
                    f"Publish with a safe bundle_version (suggested: '{safe_ver}')."
                )

            dest = (self.bundles_dir / f"{bid}@{bver}.flow").resolve()
            if dest.exists() and not overwrite:
                raise WorkflowBundleRegistryError(f"Bundle already installed: {dest.name}")

            try:
                tmp_path.replace(dest)
            except Exception as e:
                raise WorkflowBundleRegistryError(f"Failed installing bundle to {dest}: {e}") from e

            try:
                bundle2 = open_workflow_bundle(dest)
                sha = _sha256_file(dest)
                return InstalledWorkflowBundle(
                    bundle_id=str(bundle2.manifest.bundle_id),
                    bundle_version=str(bundle2.manifest.bundle_version),
                    path=dest,
                    manifest=bundle2.manifest,
                    sha256=sha,
                )
            except Exception:
                return InstalledWorkflowBundle(bundle_id=bid, bundle_version=bver, path=dest, manifest=man)
        finally:
            try:
                if tmp_path.exists():
                    tmp_path.unlink()
            except Exception:
                pass

    def remove(self, bundle_ref: str) -> int:
        """Remove installed bundle(s).

        Args:
            bundle_ref: `bundle_id` removes all versions, `bundle_id@version` removes that version.

        Returns:
            Number of removed files.
        """
        bid, ver = _split_bundle_ref(bundle_ref)
        if not bid:
            raise WorkflowBundleRegistryError("bundle_ref must be 'bundle_id' or 'bundle_id@version'")

        bundles = self.bundles_by_id().get(bid) or {}
        if not bundles:
            return 0

        removed = 0
        if ver:
            target = bundles.get(ver)
            if target is None:
                return 0
            try:
                target.path.unlink()
            except Exception as e:
                raise WorkflowBundleRegistryError(f"Failed removing {target.path}: {e}") from e
            return 1

        for b in bundles.values():
            try:
                b.path.unlink()
                removed += 1
            except Exception as e:
                raise WorkflowBundleRegistryError(f"Failed removing {b.path}: {e}") from e
        return removed

    @staticmethod
    def _entrypoint_from_entrypoint(
        *,
        ep: WorkflowBundleEntrypoint,
        bundle_id: str,
        bundle_version: str,
        is_default: bool,
    ) -> WorkflowEntrypointRef:
        fid = str(getattr(ep, "flow_id", "") or "").strip()
        name = str(getattr(ep, "name", "") or "").strip()
        desc = str(getattr(ep, "description", "") or "")
        interfaces = [str(x).strip() for x in list(getattr(ep, "interfaces", None) or []) if isinstance(x, str) and x.strip()]
        return WorkflowEntrypointRef(
            bundle_id=bundle_id,
            bundle_version=bundle_version,
            flow_id=fid,
            name=name,
            description=desc,
            interfaces=interfaces,
            is_default=bool(is_default),
        )

    def _entrypoint_from_manifest(
        self,
        *,
        manifest: WorkflowBundleManifest,
        bundle_id: str,
        bundle_version: str,
        flow_id: Optional[str],
        interface: Optional[str],
    ) -> WorkflowEntrypointRef:
        eps = list(getattr(manifest, "entrypoints", None) or [])
        if not eps:
            raise WorkflowBundleRegistryError(f"Bundle '{bundle_id}@{bundle_version}' has no entrypoints")

        default_fid = str(getattr(manifest, "default_entrypoint", "") or "").strip()
        chosen = None
        if flow_id:
            chosen = next((ep for ep in eps if str(getattr(ep, "flow_id", "") or "").strip() == flow_id), None)
            if chosen is None:
                available = ", ".join(sorted({str(getattr(ep, 'flow_id', '') or '').strip() for ep in eps if str(getattr(ep, 'flow_id', '') or '').strip()}))
                raise WorkflowBundleRegistryError(
                    f"Entrypoint '{flow_id}' not found in bundle '{bundle_id}@{bundle_version}' (available: {available})"
                )
        elif default_fid:
            chosen = next((ep for ep in eps if str(getattr(ep, "flow_id", "") or "").strip() == default_fid), None)

        if chosen is None:
            chosen = eps[0]

        epr = self._entrypoint_from_entrypoint(
            ep=chosen,
            bundle_id=bundle_id,
            bundle_version=bundle_version,
            is_default=bool(default_fid and str(getattr(chosen, "flow_id", "") or "").strip() == default_fid),
        )
        if interface and interface not in epr.interfaces:
            raise WorkflowBundleRegistryError(
                f"Selected entrypoint '{epr.workflow_id}' does not implement interface '{interface}'"
            )
        return epr



def test_llm_call_use_context_appends_turn_into_active_context():
    from abstractruntime.core.models import Effect, EffectType, RunState
    from abstractruntime.integrations.abstractcore.effect_handlers import make_llm_call_handler

    class DummyLLM:
        def generate(self, *, prompt, messages, system_prompt, media, tools, params):
            return {"content": "OK"}

    handler = make_llm_call_handler(llm=DummyLLM())

    run = RunState.new(
        workflow_id="wf",
        entry_node="node-1",
        vars={"context": {}},
    )
    run.current_node = "node-1"

    effect = Effect(
        type=EffectType.LLM_CALL,
        payload={
            "prompt": "Hello",
            "include_context": True,
            "params": {},
        },
        result_key="_temp.llm",
    )

    outcome = handler(run, effect, None)
    assert outcome.status == "completed"

    ctx = run.vars.get("context")
    assert isinstance(ctx, dict)
    messages = ctx.get("messages")
    assert isinstance(messages, list)
    assert [m.get("role") for m in messages] == ["user", "assistant"]
    assert messages[0].get("content") == "Hello"
    assert messages[1].get("content") == "OK"



"""abstractruntime.integrations.abstractcore.summarizer

Integration with AbstractCore's BasicSummarizer for chat compaction.

This module provides a wrapper around AbstractCore's BasicSummarizer
that respects environment token limits (max_tokens, max_output_tokens)
for adaptive chunking during conversation compaction.

Design:
- The kernel (runtime.py) uses this via dependency injection
- BasicSummarizer handles adaptive chunking based on max_tokens
- When max_tokens == -1: Uses model's full capability (AUTO mode)
- When max_tokens > 0: Uses explicit limit (environment constraint)
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Protocol


class ChatSummarizer(Protocol):
    """Protocol for chat history summarization.

    This protocol allows the runtime kernel to use summarization
    without directly importing AbstractCore.
    """

    def summarize_chat_history(
        self,
        messages: List[Dict[str, Any]],
        *,
        preserve_recent: int = 6,
        focus: Optional[str] = None,
        compression_mode: str = "standard",
    ) -> Dict[str, Any]:
        """Summarize chat history with adaptive chunking.

        Args:
            messages: List of message dicts with 'role' and 'content' keys
            preserve_recent: Number of recent messages to keep intact (default 6)
            focus: Optional focus for summarization
            compression_mode: How aggressively to compress (light|standard|heavy)

        Returns:
            Dict with keys: summary, key_points, confidence, focus_alignment
        """
        ...


class AbstractCoreChatSummarizer:
    """Wrapper around AbstractCore's BasicSummarizer for runtime integration.

    This class:
    - Wraps BasicSummarizer with token limits from RuntimeConfig
    - Handles adaptive chunking automatically via BasicSummarizer
    - Returns JSON-safe dicts for storage in RunState.vars

    Example:
        >>> summarizer = AbstractCoreChatSummarizer(
        ...     llm=llm_instance,
        ...     max_tokens=32768,
        ...     max_output_tokens=4096,
        ... )
        >>> result = summarizer.summarize_chat_history(messages)
        >>> print(result["summary"])
    """

    def __init__(
        self,
        llm,
        *,
        max_tokens: int = -1,
        max_output_tokens: int = -1,
    ):
        """Initialize the summarizer with token limits.

        Args:
            llm: AbstractCore LLM instance (from create_llm or provider)
            max_tokens: Maximum context tokens. -1 = AUTO (use model capability)
            max_output_tokens: Maximum output tokens. -1 = AUTO
        """
        from abstractcore.processing import BasicSummarizer

        self._summarizer = BasicSummarizer(
            llm=llm,
            max_tokens=max_tokens,
            max_output_tokens=max_output_tokens,
        )
        self._max_tokens = max_tokens
        self._max_output_tokens = max_output_tokens

    def summarize_chat_history(
        self,
        messages: List[Dict[str, Any]],
        *,
        preserve_recent: int = 6,
        focus: Optional[str] = None,
        compression_mode: str = "standard",
    ) -> Dict[str, Any]:
        """Summarize chat history with adaptive chunking.

        When max_tokens > 0 and messages exceed the limit, BasicSummarizer
        automatically uses map-reduce chunking to process the content.

        Args:
            messages: List of message dicts with 'role' and 'content' keys
            preserve_recent: Number of recent messages to keep intact (default 6)
            focus: Optional focus for summarization (e.g., "key decisions")
            compression_mode: How aggressively to compress (light|standard|heavy)

        Returns:
            Dict with keys:
                - summary: The summarized text
                - key_points: List of key points extracted
                - confidence: Confidence score (0-1)
                - focus_alignment: How well summary addresses focus (0-1)
        """
        from abstractcore.processing import CompressionMode

        # Map string to enum
        mode_map = {
            "light": CompressionMode.LIGHT,
            "standard": CompressionMode.STANDARD,
            "heavy": CompressionMode.HEAVY,
        }
        mode = mode_map.get(compression_mode.lower(), CompressionMode.STANDARD)

        # Call BasicSummarizer - it handles adaptive chunking internally
        result = self._summarizer.summarize_chat_history(
            messages=messages,
            preserve_recent=preserve_recent,
            focus=focus,
            compression_mode=mode,
        )

        # Return as JSON-safe dict for storage in RunState.vars
        return {
            "summary": result.summary,
            "key_points": list(result.key_points) if result.key_points else [],
            "confidence": float(result.confidence) if result.confidence else None,
            "focus_alignment": float(result.focus_alignment) if result.focus_alignment else None,
            "word_count_original": result.word_count_original,
            "word_count_summary": result.word_count_summary,
        }

    @property
    def max_tokens(self) -> int:
        """Current max_tokens setting."""
        return self._max_tokens

    @property
    def max_output_tokens(self) -> int:
        """Current max_output_tokens setting."""
        return self._max_output_tokens



from __future__ import annotations

import tempfile
from pathlib import Path

from abstractruntime.storage.commands import (
    CommandRecord,
    InMemoryCommandCursorStore,
    InMemoryCommandStore,
    JsonFileCommandCursorStore,
    JsonlCommandStore,
)
from abstractruntime.storage.sqlite import SqliteCommandCursorStore, SqliteCommandStore, SqliteDatabase


def test_in_memory_command_store_idempotency_and_cursor() -> None:
    store = InMemoryCommandStore()

    r1 = store.append(CommandRecord(command_id="c1", run_id="run_1", type="pause", payload={}, ts="t", seq=0))
    assert r1.accepted is True
    assert r1.duplicate is False
    assert r1.seq == 1

    r2 = store.append(CommandRecord(command_id="c1", run_id="run_1", type="pause", payload={}, ts="t", seq=0))
    assert r2.accepted is False
    assert r2.duplicate is True
    assert r2.seq == 1

    items, cur = store.list_after(after=0, limit=10)
    assert [x.command_id for x in items] == ["c1"]
    assert cur == 1

    items2, cur2 = store.list_after(after=1, limit=10)
    assert items2 == []
    assert cur2 == 1


def test_jsonl_command_store_idempotency_and_restart_cursor() -> None:
    with tempfile.TemporaryDirectory(prefix="abstractruntime-cmdstore-") as td:
        base = Path(td)
        store = JsonlCommandStore(base)

        r1 = store.append(CommandRecord(command_id="c1", run_id="run_1", type="pause", payload={}, ts="t", seq=0))
        r2 = store.append(CommandRecord(command_id="c2", run_id="run_1", type="cancel", payload={}, ts="t", seq=0))
        assert r1.seq == 1
        assert r2.seq == 2

        r1b = store.append(CommandRecord(command_id="c1", run_id="run_1", type="pause", payload={}, ts="t", seq=0))
        assert r1b.duplicate is True
        assert store.get_last_seq() == 2

        items, cur = store.list_after(after=0, limit=10)
        assert [x.command_id for x in items] == ["c1", "c2"]
        assert cur == 2

        # Restart: re-open from disk and confirm idempotency holds.
        store2 = JsonlCommandStore(base)
        r2b = store2.append(CommandRecord(command_id="c2", run_id="run_1", type="cancel", payload={}, ts="t", seq=0))
        assert r2b.duplicate is True
        assert store2.get_last_seq() == 2

        items2, cur2 = store2.list_after(after=1, limit=10)
        assert [x.command_id for x in items2] == ["c2"]
        assert cur2 == 2


def test_command_cursor_store_roundtrip() -> None:
    mem = InMemoryCommandCursorStore()
    assert mem.load() == 0
    mem.save(12)
    assert mem.load() == 12

    with tempfile.TemporaryDirectory(prefix="abstractruntime-cursor-") as td:
        p = Path(td) / "cursor.json"
        fs = JsonFileCommandCursorStore(p)
        assert fs.load() == 0
        fs.save(7)
        assert fs.load() == 7


def test_sqlite_command_store_idempotency_and_restart_cursor(tmp_path: Path) -> None:
    db_path = tmp_path / "gateway.sqlite3"
    db = SqliteDatabase(db_path)
    store = SqliteCommandStore(db)
    cursor_store = SqliteCommandCursorStore(db, consumer_id="t")

    r1 = store.append(CommandRecord(command_id="c1", run_id="run_1", type="pause", payload={}, ts="t", seq=0))
    assert r1.accepted is True
    assert r1.duplicate is False
    assert r1.seq >= 1

    r1b = store.append(CommandRecord(command_id="c1", run_id="run_1", type="pause", payload={}, ts="t", seq=0))
    assert r1b.accepted is False
    assert r1b.duplicate is True
    assert r1b.seq == r1.seq

    cursor_store.save(r1.seq)
    assert cursor_store.load() == r1.seq

    # Restart: re-open and verify both command idempotency and cursor persist.
    db2 = SqliteDatabase(db_path)
    store2 = SqliteCommandStore(db2)
    cursor2 = SqliteCommandCursorStore(db2, consumer_id="t")
    assert cursor2.load() == r1.seq

    r2 = store2.append(CommandRecord(command_id="c2", run_id="run_1", type="cancel", payload={}, ts="t", seq=0))
    assert r2.accepted is True
    assert store2.get_last_seq() >= r2.seq

    items, cur = store2.list_after(after=0, limit=10)
    assert [x.command_id for x in items] == ["c1", "c2"]
    assert cur == items[-1].seq




import json
import sys
from pathlib import Path

import pytest


def _repo_root() -> Path:
    # .../abstractruntime/tests/<file> -> .../abstractframework
    return Path(__file__).resolve().parents[2]


def test_visualflow_compiler_compiles_real_flow_without_abstractflow_import() -> None:
    before = set(sys.modules.keys())

    from abstractruntime.visualflow_compiler import compile_visualflow

    path = (
        _repo_root()
        / "docs"
        / "assessments"
        / "examples"
        / "workflowartifact-abstraction"
        / "ac-advanced-agent.visualflow.json"
    )
    if not path.exists():
        pytest.skip("VisualFlow fixture not present under docs/assessments (optional in some workspaces)")
    raw = json.loads(path.read_text(encoding="utf-8"))

    spec = compile_visualflow(raw)
    assert spec.workflow_id == "ac-advanced-agent"
    assert spec.entry_node == "node-1"
    assert isinstance(spec.nodes, dict)
    assert len(spec.nodes) == 4

    # The compiler must not import AbstractFlow (single semantics engine lives in runtime).
    after = set(sys.modules.keys())
    newly_loaded = after - before
    assert not any(m == "abstractflow" or m.startswith("abstractflow.") for m in newly_loaded)


def test_visualflow_compiler_accepts_model_dump_like_objects() -> None:
    from abstractruntime.visualflow_compiler import compile_visualflow

    path = (
        _repo_root()
        / "docs"
        / "assessments"
        / "examples"
        / "workflowartifact-abstraction"
        / "ac-advanced-agent.visualflow.json"
    )
    if not path.exists():
        pytest.skip("VisualFlow fixture not present under docs/assessments (optional in some workspaces)")
    raw = json.loads(path.read_text(encoding="utf-8"))

    class Dummy:
        def model_dump(self):
            return raw

    spec = compile_visualflow(Dummy())
    assert spec.workflow_id == "ac-advanced-agent"


def test_visualflow_compiler_tree_compiles_bundle_flows() -> None:
    before = set(sys.modules.keys())

    from abstractruntime.visualflow_compiler import compile_visualflow_tree

    flows_dir = (
        _repo_root()
        / "docs"
        / "assessments"
        / "examples"
        / "workflowartifact-abstraction"
        / "ac-advanced-agent.bundle"
        / "flows"
    )
    if not flows_dir.exists():
        pytest.skip("VisualFlow bundle fixtures not present under docs/assessments (optional in some workspaces)")
    flows = {p.stem: json.loads(p.read_text(encoding="utf-8")) for p in flows_dir.glob("*.json")}

    specs = compile_visualflow_tree(root_id="ac-advanced-agent", flows_by_id=flows)
    assert set(specs.keys()) == {"ac-advanced-agent", "4ed3b340", "60a97e4d", "15f19f7f"}
    after = set(sys.modules.keys())
    newly_loaded = after - before
    assert not any(m == "abstractflow" or m.startswith("abstractflow.") for m in newly_loaded)



from __future__ import annotations


def test_visualflow_llm_call_context_attachments_are_forwarded_as_media() -> None:
    """VisualFlow LLM Call should map `context.attachments` -> `payload.media`."""
    from abstractruntime.core.models import RunState
    from abstractruntime.visualflow_compiler import compile_visualflow

    spec = compile_visualflow(
        {
            "id": "vf",
            "name": "vf",
            "entryNode": "call",
            "nodes": [
                {
                    "id": "call",
                    "type": "llm_call",
                    "data": {"effectConfig": {"provider": "lmstudio", "model": "unit-test-model"}},
                }
            ],
            "edges": [],
        }
    )

    run = RunState.new(
        workflow_id=spec.workflow_id,
        entry_node="call",
        vars={
            "_last_output": {
                "prompt": "Hello",
                "context": {
                    "attachments": [{"$artifact": "a1", "filename": "notes.txt"}],
                    "messages": [{"role": "system", "content": "ctx-system"}],
                },
            }
        },
    )

    plan = spec.get_node("call")(run, None)
    assert plan.effect is not None
    payload = dict(plan.effect.payload or {})
    assert payload.get("media") == [{"$artifact": "a1", "filename": "notes.txt"}]




def test_visual_agent_unstructured_result_is_minimal_and_messages_live_in_scratchpad():
    """Agent unstructured output should avoid duplicating transcript/context in `result`.

    The user-facing answer is surfaced on the `response` output pin, while the agent-internal
    transcript is stored under the scratchpad (runtime-owned).
    """
    from abstractruntime.core.models import RunState
    from abstractruntime.visualflow_compiler.compiler import compile_visualflow

    spec = compile_visualflow(
        {
            "id": "test-flow",
            "name": "test",
            "nodes": [
                {
                    "id": "node-agent",
                    "type": "agent",
                    "data": {
                        "agentConfig": {},
                    },
                }
            ],
            "edges": [],
            "entryNode": "node-agent",
        }
    )

    handler = spec.nodes["node-agent"]

    run = RunState.new(
        workflow_id=spec.workflow_id,
        entry_node="node-agent",
        vars={"context": {}},
    )

    run.vars["_temp"] = {
        "agent": {
            "node-agent": {
                "phase": "subworkflow",
                "resolved_inputs": {
                    "task": "Do the thing.",
                    "context": {"foo": "bar"},
                    "provider": "lmstudio",
                    "model": "dummy",
                    "include_context": False,
                    "tools": [],
                },
                "sub": {
                    "sub_run_id": "sub-1",
                    "output": {
                        "answer": "Done.",
                        "iterations": 1,
                        "messages": [
                            {"role": "user", "content": "Do the thing.", "metadata": {"message_id": "msg_u"}},
                            {"role": "assistant", "content": "Done.", "metadata": {"message_id": "msg_a"}},
                        ],
                    },
                    "node_traces": {},
                },
            }
        }
    }

    handler(run, None)

    temp = run.vars.get("_temp")
    assert isinstance(temp, dict)
    effects = temp.get("effects")
    assert isinstance(effects, dict)
    result = effects.get("node-agent")
    assert isinstance(result, dict)

    assert result.get("success") is True
    assert result.get("context") is None
    assert result.get("messages") is None

    agent_ns = temp.get("agent")
    assert isinstance(agent_ns, dict)
    bucket = agent_ns.get("node-agent")
    assert isinstance(bucket, dict)
    scratchpad = bucket.get("scratchpad")
    assert isinstance(scratchpad, dict)
    assert scratchpad.get("task") == "Do the thing."
    assert scratchpad.get("context_extra") == {"foo": "bar"}
    messages = scratchpad.get("messages")
    assert isinstance(messages, list)
    assert [m.get("role") for m in messages] == ["user", "assistant"]
    assert messages[0].get("content") == "Do the thing."
    assert messages[1].get("content") == "Done."


def test_visual_agent_structured_result_sync_exposes_data_pin():
    from abstractruntime.core.models import RunState
    from abstractruntime.visualflow_compiler.compiler import _sync_effect_results_to_node_outputs
    from abstractruntime.visualflow_compiler.visual.executor import visual_to_flow
    from abstractruntime.visualflow_compiler.visual.models import load_visualflow_json

    vf = load_visualflow_json(
        {
            "id": "vf-agent-structured-sync",
            "name": "vf-agent-structured-sync",
            "entryNode": "node-agent",
            "nodes": [
                {
                    "id": "node-agent",
                    "type": "agent",
                    "data": {
                        "nodeType": "agent",
                        "pinDefaults": {
                            "resp_schema": {
                                "type": "object",
                                "properties": {
                                    "choice": {"type": "string", "enum": ["summarize", "classify"]},
                                    "confidence": {"type": "number"},
                                },
                                "required": ["choice"],
                            }
                        },
                        "outputs": [
                            {"id": "exec-out", "label": "", "type": "execution"},
                            {"id": "response", "label": "response", "type": "string"},
                            {"id": "data", "label": "data", "type": "object"},
                            {"id": "meta", "label": "meta", "type": "object"},
                            {"id": "scratchpad", "label": "scratchpad", "type": "object"},
                        ],
                    },
                }
            ],
            "edges": [],
        }
    )
    flow = visual_to_flow(vf)
    run = RunState.new(workflow_id=flow.flow_id, entry_node="node-agent")
    structured_data = {"choice": "summarize", "confidence": 0.95}
    run.vars["_temp"] = {
        "effects": {"node-agent": structured_data},
        "agent": {
            "node-agent": {
                "output_mode": "structured",
                "answer": "The user asked for a summary.",
                "provider": "lmstudio",
                "model": "qwen/qwen3.6-35b-a3b",
                "scratchpad": {
                    "node_id": "node-agent",
                    "steps": [],
                    "messages": [{"role": "assistant", "content": "The user asked for a summary."}],
                },
            }
        },
    }

    _sync_effect_results_to_node_outputs(run, flow)

    out = flow._node_outputs["node-agent"]  # type: ignore[attr-defined]
    assert out["response"] == '{"choice":"summarize","confidence":0.95}'
    assert out["data"] == structured_data
    assert out["meta"]["output_mode"] == "structured"
    assert out["meta"]["provider"] == "lmstudio"
    assert out["meta"]["model"] == "qwen/qwen3.6-35b-a3b"



from __future__ import annotations

from abstractruntime import Effect, EffectType, Runtime, StepPlan, WorkflowSpec
from abstractruntime.core.runtime import EffectOutcome
from abstractruntime.storage.in_memory import InMemoryLedgerStore, InMemoryRunStore


def test_llm_effect_trace_metadata_includes_runtime_step_identity() -> None:
    seen_payloads = []

    def call_node(_run, _ctx):
        return StepPlan(
            node_id="call",
            effect=Effect(
                type=EffectType.LLM_CALL,
                payload={"prompt": "hello", "params": {}},
                result_key="_temp.effects.call",
            ),
        )

    def llm_handler(_run, effect, _default_next_node):
        seen_payloads.append(effect.payload)
        return EffectOutcome.completed({"content": "ok"})

    run_store = InMemoryRunStore()
    ledger_store = InMemoryLedgerStore()
    runtime = Runtime(
        run_store=run_store,
        ledger_store=ledger_store,
        effect_handlers={EffectType.LLM_CALL: llm_handler},
    )
    workflow = WorkflowSpec(workflow_id="wf-step-trace", entry_node="call", nodes={"call": call_node})

    run_id = runtime.start(workflow=workflow)
    runtime.tick(workflow=workflow, run_id=run_id)

    started = [r for r in ledger_store.list(run_id) if r.get("status") == "started"]
    assert len(started) == 1
    step_id = started[0]["step_id"]
    idempotency_key = started[0]["idempotency_key"]
    trace = started[0]["effect"]["payload"]["params"]["trace_metadata"]
    assert trace["step_id"] == step_id
    assert trace["effect_idempotency_key"] == idempotency_key
    assert trace["attempt"] == 1
    assert seen_payloads[0]["params"]["trace_metadata"]["step_id"] == step_id



"""Public host-facing AbstractCore control facade for Runtime integrations.

Hosts should import this module instead of reaching through
`runtime._abstractcore_llm_client` directly.

Scope:
- prompt-cache control operations
- host-local prompt-cache export/import admin operations
- durable bloc/KV prompt-cache operations
- durable bloc/KV lifecycle operations
- model-residency control operations
- host-local email/comms helper operations

Non-goals:
- durable Runtime effect execution (use `AbstractCoreRunFacade` / `get_abstractcore_run_facade(...)`)
- generated media / TTS / STT helpers
- provider-private prompt-cache export/import behavior
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Protocol

_RUNTIME_ABSTRACTCORE_CLIENT_ATTR = "_abstractcore_llm_client"
_HOST_CONTROL_METHODS = (
    "get_prompt_cache_capabilities",
    "get_prompt_cache_stats",
    "prompt_cache_set",
    "prompt_cache_update",
    "prompt_cache_fork",
    "prompt_cache_clear",
    "prompt_cache_prepare_modules",
    "list_prompt_cache_exports",
    "prompt_cache_export",
    "prompt_cache_import",
    "upsert_text_bloc",
    "get_bloc_record",
    "list_blocs",
    "get_bloc_kv_manifest",
    "ensure_bloc_kv_artifact",
    "load_bloc_kv_artifact",
    "list_bloc_kv_artifacts",
    "delete_bloc_kv_artifact",
    "prune_bloc_kv_artifacts",
    "delete_bloc",
    "get_model_residency_capabilities",
    "list_model_residency",
    "load_model_residency",
    "unload_model_residency",
)


class AbstractCoreHostControlClient(Protocol):
    """Duck-typed control client contract implemented by AbstractCore LLM clients."""

    def get_prompt_cache_capabilities(self, **kwargs: Any) -> Dict[str, Any]:
        ...

    def get_prompt_cache_stats(self, **kwargs: Any) -> Dict[str, Any]:
        ...

    def prompt_cache_set(
        self,
        *,
        key: str,
        make_default: bool = True,
        ttl_s: Optional[float] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def prompt_cache_update(
        self,
        *,
        key: str,
        prompt: Optional[str] = None,
        messages: Optional[List[Dict[str, Any]]] = None,
        system_prompt: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        add_generation_prompt: bool = False,
        ttl_s: Optional[float] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def prompt_cache_fork(
        self,
        *,
        from_key: str,
        to_key: str,
        make_default: bool = False,
        ttl_s: Optional[float] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def prompt_cache_clear(
        self,
        *,
        key: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def prompt_cache_prepare_modules(
        self,
        *,
        namespace: str,
        modules: List[Dict[str, Any]],
        make_default: bool = False,
        ttl_s: Optional[float] = None,
        version: int = 1,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def list_prompt_cache_exports(
        self,
        *,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def prompt_cache_export(
        self,
        *,
        name: str,
        key: str,
        q8: bool = False,
        meta: Optional[Dict[str, Any]] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def prompt_cache_import(
        self,
        *,
        name: str,
        key: Optional[str] = None,
        make_default: bool = True,
        clear_existing: bool = False,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def upsert_text_bloc(
        self,
        *,
        path: str,
        content: str,
        sha256: Optional[str] = None,
        content_sha256: Optional[str] = None,
        media_type: str = "text",
        size_bytes: Optional[int] = None,
        mtime_ns: Optional[int] = None,
        format: Optional[str] = None,
        estimated_tokens: Optional[int] = None,
        relpath_base: Optional[str] = None,
        summary: Optional[str] = None,
        keywords: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def get_bloc_record(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def list_blocs(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def get_bloc_kv_manifest(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        artifact_path: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def ensure_bloc_kv_artifact(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        artifact_path: Optional[str] = None,
        force_rebuild: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def load_bloc_kv_artifact(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        artifact_path: Optional[str] = None,
        stable_cache_key: Optional[str] = None,
        key: Optional[str] = None,
        make_default: bool = False,
        force_rebuild: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def list_bloc_kv_artifacts(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def delete_bloc_kv_artifact(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        artifact_path: Optional[str] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        clear_loaded: bool = False,
        force: bool = False,
        dry_run: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def prune_bloc_kv_artifacts(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        clear_loaded: bool = False,
        force: bool = False,
        dry_run: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def delete_bloc(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        delete_kv: bool = True,
        clear_loaded: bool = False,
        force: bool = False,
        dry_run: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def get_model_residency_capabilities(self, **kwargs: Any) -> Dict[str, Any]:
        ...

    def list_model_residency(
        self,
        *,
        task: Optional[str] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def load_model_residency(
        self,
        *,
        task: str,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        pin: bool = True,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    def unload_model_residency(
        self,
        *,
        task: Optional[str] = None,
        runtime_id: Optional[str] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...


def _coerce_control_client(client: Any, *, source: str) -> AbstractCoreHostControlClient:
    missing = [name for name in _HOST_CONTROL_METHODS if not callable(getattr(client, name, None))]
    if missing:
        methods = ", ".join(missing)
        raise TypeError(
            f"{source} does not implement the AbstractCore host control contract. Missing methods: {methods}."
        )
    return client


def _client_from_runtime(runtime: Any) -> AbstractCoreHostControlClient:
    client = getattr(runtime, _RUNTIME_ABSTRACTCORE_CLIENT_ATTR, None)
    if client is None:
        raise RuntimeError(
            "Runtime is not wired to AbstractCore host controls. "
            "Use an AbstractCore runtime factory, then obtain the facade via "
            "`abstractruntime.integrations.abstractcore.get_abstractcore_host_facade(runtime)`."
        )
    return _coerce_control_client(client, source="Runtime AbstractCore client")


class AbstractCoreHostFacade:
    """Public host-facing control surface for AbstractCore-backed runtimes.

    This facade is intentionally narrow and operator-scoped. Its methods are
    synchronous helpers and do not create durable Runtime history on their own.
    Most methods delegate through the configured Runtime client; the small
    comms/email helpers intentionally remain host-local wrappers over
    AbstractCore's public tool modules in phase 1.
    """

    __slots__ = ("_client",)

    def __init__(self, runtime: Any):
        self._client = _client_from_runtime(runtime)

    @classmethod
    def from_runtime(cls, runtime: Any) -> "AbstractCoreHostFacade":
        """Bind the public host facade to an AbstractCore-wired runtime."""

        return cls(runtime)

    def get_prompt_cache_capabilities(self, **kwargs: Any) -> Dict[str, Any]:
        return self._client.get_prompt_cache_capabilities(**kwargs)

    def get_prompt_cache_stats(self, **kwargs: Any) -> Dict[str, Any]:
        return self._client.get_prompt_cache_stats(**kwargs)

    def get_model_residency_capabilities(self, **kwargs: Any) -> Dict[str, Any]:
        return self._client.get_model_residency_capabilities(**kwargs)

    def prompt_cache_set(
        self,
        *,
        key: str,
        make_default: bool = True,
        ttl_s: Optional[float] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.prompt_cache_set(
            key=key,
            make_default=make_default,
            ttl_s=ttl_s,
            **kwargs,
        )

    def prompt_cache_update(
        self,
        *,
        key: str,
        prompt: Optional[str] = None,
        messages: Optional[List[Dict[str, Any]]] = None,
        system_prompt: Optional[str] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        add_generation_prompt: bool = False,
        ttl_s: Optional[float] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.prompt_cache_update(
            key=key,
            prompt=prompt,
            messages=messages,
            system_prompt=system_prompt,
            tools=tools,
            add_generation_prompt=add_generation_prompt,
            ttl_s=ttl_s,
            **kwargs,
        )

    def prompt_cache_fork(
        self,
        *,
        from_key: str,
        to_key: str,
        make_default: bool = False,
        ttl_s: Optional[float] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.prompt_cache_fork(
            from_key=from_key,
            to_key=to_key,
            make_default=make_default,
            ttl_s=ttl_s,
            **kwargs,
        )

    def prompt_cache_clear(
        self,
        *,
        key: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.prompt_cache_clear(key=key, **kwargs)

    def prompt_cache_prepare_modules(
        self,
        *,
        namespace: str,
        modules: List[Dict[str, Any]],
        make_default: bool = False,
        ttl_s: Optional[float] = None,
        version: int = 1,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.prompt_cache_prepare_modules(
            namespace=namespace,
            modules=modules,
            make_default=make_default,
            ttl_s=ttl_s,
            version=version,
            **kwargs,
        )

    def list_prompt_cache_exports(
        self,
        *,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.list_prompt_cache_exports(provider=provider, model=model, **kwargs)

    def prompt_cache_export(
        self,
        *,
        name: str,
        key: str,
        q8: bool = False,
        meta: Optional[Dict[str, Any]] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.prompt_cache_export(
            name=name,
            key=key,
            q8=q8,
            meta=meta,
            provider=provider,
            model=model,
            **kwargs,
        )

    def prompt_cache_import(
        self,
        *,
        name: str,
        key: Optional[str] = None,
        make_default: bool = True,
        clear_existing: bool = False,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.prompt_cache_import(
            name=name,
            key=key,
            make_default=make_default,
            clear_existing=clear_existing,
            provider=provider,
            model=model,
            **kwargs,
        )

    def upsert_text_bloc(
        self,
        *,
        path: str,
        content: str,
        sha256: Optional[str] = None,
        content_sha256: Optional[str] = None,
        media_type: str = "text",
        size_bytes: Optional[int] = None,
        mtime_ns: Optional[int] = None,
        format: Optional[str] = None,
        estimated_tokens: Optional[int] = None,
        relpath_base: Optional[str] = None,
        summary: Optional[str] = None,
        keywords: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.upsert_text_bloc(
            path=path,
            content=content,
            sha256=sha256,
            content_sha256=content_sha256,
            media_type=media_type,
            size_bytes=size_bytes,
            mtime_ns=mtime_ns,
            format=format,
            estimated_tokens=estimated_tokens,
            relpath_base=relpath_base,
            summary=summary,
            keywords=keywords,
            **kwargs,
        )

    def get_bloc_record(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.get_bloc_record(sha256=sha256, bloc_id=bloc_id, **kwargs)

    def list_blocs(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.list_blocs(sha256=sha256, bloc_id=bloc_id, **kwargs)

    def get_bloc_kv_manifest(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        artifact_path: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.get_bloc_kv_manifest(
            sha256=sha256,
            bloc_id=bloc_id,
            artifact_path=artifact_path,
            **kwargs,
        )

    def ensure_bloc_kv_artifact(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        artifact_path: Optional[str] = None,
        force_rebuild: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.ensure_bloc_kv_artifact(
            sha256=sha256,
            bloc_id=bloc_id,
            artifact_path=artifact_path,
            force_rebuild=force_rebuild,
            debug=debug,
            **kwargs,
        )

    def load_bloc_kv_artifact(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        artifact_path: Optional[str] = None,
        stable_cache_key: Optional[str] = None,
        key: Optional[str] = None,
        make_default: bool = False,
        force_rebuild: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.load_bloc_kv_artifact(
            sha256=sha256,
            bloc_id=bloc_id,
            artifact_path=artifact_path,
            stable_cache_key=stable_cache_key,
            key=key,
            make_default=make_default,
            force_rebuild=force_rebuild,
            debug=debug,
            **kwargs,
        )

    def list_bloc_kv_artifacts(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.list_bloc_kv_artifacts(
            sha256=sha256,
            bloc_id=bloc_id,
            provider=provider,
            model=model,
            **kwargs,
        )

    def delete_bloc_kv_artifact(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        artifact_path: Optional[str] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        clear_loaded: bool = False,
        force: bool = False,
        dry_run: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.delete_bloc_kv_artifact(
            sha256=sha256,
            bloc_id=bloc_id,
            artifact_path=artifact_path,
            provider=provider,
            model=model,
            clear_loaded=clear_loaded,
            force=force,
            dry_run=dry_run,
            debug=debug,
            **kwargs,
        )

    def prune_bloc_kv_artifacts(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        clear_loaded: bool = False,
        force: bool = False,
        dry_run: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.prune_bloc_kv_artifacts(
            sha256=sha256,
            bloc_id=bloc_id,
            provider=provider,
            model=model,
            clear_loaded=clear_loaded,
            force=force,
            dry_run=dry_run,
            debug=debug,
            **kwargs,
        )

    def delete_bloc(
        self,
        *,
        sha256: Optional[str] = None,
        bloc_id: Optional[int] = None,
        delete_kv: bool = True,
        clear_loaded: bool = False,
        force: bool = False,
        dry_run: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.delete_bloc(
            sha256=sha256,
            bloc_id=bloc_id,
            delete_kv=delete_kv,
            clear_loaded=clear_loaded,
            force=force,
            dry_run=dry_run,
            **kwargs,
        )

    def list_model_residency(
        self,
        *,
        task: Optional[str] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.list_model_residency(
            task=task,
            provider=provider,
            model=model,
            **kwargs,
        )

    def load_model_residency(
        self,
        *,
        task: str,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        pin: bool = True,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.load_model_residency(
            task=task,
            provider=provider,
            model=model,
            options=options,
            pin=pin,
            **kwargs,
        )

    def unload_model_residency(
        self,
        *,
        task: Optional[str] = None,
        runtime_id: Optional[str] = None,
        provider: Optional[str] = None,
        model: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return self._client.unload_model_residency(
            task=task,
            runtime_id=runtime_id,
            provider=provider,
            model=model,
            options=options,
            **kwargs,
        )

    def list_email_accounts(self) -> Dict[str, Any]:
        from .comms_facade import list_email_accounts

        return list_email_accounts()

    def list_emails(
        self,
        *,
        account: Optional[str] = None,
        mailbox: Optional[str] = None,
        since: Optional[str] = None,
        status: str = "all",
        limit: int = 20,
        timeout_s: float = 30.0,
    ) -> Dict[str, Any]:
        from .comms_facade import list_emails

        return list_emails(
            account=account,
            mailbox=mailbox,
            since=since,
            status=status,
            limit=limit,
            timeout_s=timeout_s,
        )

    def read_email(
        self,
        *,
        uid: str,
        account: Optional[str] = None,
        mailbox: Optional[str] = None,
        timeout_s: float = 30.0,
        max_body_chars: int = 20000,
    ) -> Dict[str, Any]:
        from .comms_facade import read_email

        return read_email(
            uid=uid,
            account=account,
            mailbox=mailbox,
            timeout_s=timeout_s,
            max_body_chars=max_body_chars,
        )

    def send_email(
        self,
        to: Any,
        subject: str,
        *,
        account: Optional[str] = None,
        body_text: Optional[str] = None,
        body_html: Optional[str] = None,
        cc: Any = None,
        bcc: Any = None,
        timeout_s: float = 30.0,
        headers: Optional[Dict[str, str]] = None,
    ) -> Dict[str, Any]:
        """Host-local operator send helper.

        If the outbound send belongs to a workflow/run, use
        `get_abstractcore_run_facade(runtime).send_email(...)` instead so
        Runtime authors the durable child-run truth.
        """
        from .comms_facade import send_email

        return send_email(
            to=to,
            subject=subject,
            account=account,
            body_text=body_text,
            body_html=body_html,
            cc=cc,
            bcc=bcc,
            timeout_s=timeout_s,
            headers=headers,
        )


def get_abstractcore_host_facade(runtime: Any) -> AbstractCoreHostFacade:
    """Return the public AbstractCore host facade bound to a runtime."""

    return AbstractCoreHostFacade.from_runtime(runtime)


__all__ = [
    "AbstractCoreHostFacade",
    "get_abstractcore_host_facade",
]



from __future__ import annotations

from pathlib import Path

import pytest

from abstractruntime.utils.workspace_paths import build_workspace_mounts


def test_workspace_or_allowed_resolves_mount_prefixed_relative_paths(tmp_path: Path) -> None:
    from abstractruntime.integrations.abstractcore.workspace_scoped_tools import WorkspaceScope, resolve_user_path

    ws = tmp_path / "ws"
    ws.mkdir(parents=True, exist_ok=True)

    allowed = tmp_path / "mnemosyne"
    allowed.mkdir(parents=True, exist_ok=True)
    (allowed / "file.txt").write_text("hi", encoding="utf-8")

    blocked = allowed / "blocked"
    blocked.mkdir(parents=True, exist_ok=True)

    scope = WorkspaceScope.from_input_data(
        {
            "workspace_root": str(ws),
            "workspace_access_mode": "workspace_or_allowed",
            "workspace_allowed_paths": [str(allowed)],
            "workspace_ignored_paths": [str(blocked)],
        }
    )
    assert scope is not None

    resolved = resolve_user_path(scope=scope, user_path="mnemosyne/file.txt")
    assert resolved == (allowed / "file.txt").resolve()

    with pytest.raises(ValueError, match="Path escapes workspace_root"):
        resolve_user_path(scope=scope, user_path="mnemosyne/../ws/oops.txt")

    with pytest.raises(ValueError, match="workspace_ignored_paths"):
        resolve_user_path(scope=scope, user_path="mnemosyne/blocked/x.txt")


def test_redundant_workspace_dir_prefix_is_stripped_for_relative_paths(tmp_path: Path) -> None:
    from abstractruntime.integrations.abstractcore.workspace_scoped_tools import WorkspaceScope, resolve_user_path

    ws = tmp_path / "mnemosyne"
    (ws / "memory").mkdir(parents=True, exist_ok=True)
    (ws / "memory" / "Index.md").write_text("hello", encoding="utf-8")

    scope = WorkspaceScope.from_input_data({"workspace_root": str(ws), "workspace_access_mode": "workspace_only"})
    assert scope is not None

    resolved = resolve_user_path(scope=scope, user_path="mnemosyne/memory/Index.md")
    assert resolved == (ws / "memory" / "Index.md").resolve()

    resolved2 = resolve_user_path(scope=scope, user_path="@mnemosyne/memory/Index.md")
    assert resolved2 == (ws / "memory" / "Index.md").resolve()


def test_mount_alias_collisions_use_shared_digest_aliases(tmp_path: Path) -> None:
    from abstractruntime.integrations.abstractcore.workspace_scoped_tools import WorkspaceScope, resolve_user_path

    ws = tmp_path / "workspace"
    ws.mkdir(parents=True, exist_ok=True)

    first = tmp_path / "team-a" / "reports"
    second = tmp_path / "team-b" / "reports"
    first.mkdir(parents=True, exist_ok=True)
    second.mkdir(parents=True, exist_ok=True)
    (first / "alpha.txt").write_text("alpha", encoding="utf-8")
    (second / "beta.txt").write_text("beta", encoding="utf-8")

    mounts = build_workspace_mounts(allowed_dirs=[first, second], used_names=set())
    assert len(mounts) == 2
    assert all(name.startswith("reports_") for name in mounts.keys())

    scope = WorkspaceScope.from_input_data(
        {
            "workspace_root": str(ws),
            "workspace_access_mode": "workspace_or_allowed",
            "workspace_allowed_paths": [str(first), str(second)],
        }
    )
    assert scope is not None

    first_alias = next(name for name, path in mounts.items() if path == first.resolve())
    second_alias = next(name for name, path in mounts.items() if path == second.resolve())
    assert resolve_user_path(scope=scope, user_path=f"{first_alias}/alpha.txt") == (first / "alpha.txt").resolve()
    assert resolve_user_path(scope=scope, user_path=f"{second_alias}/beta.txt") == (second / "beta.txt").resolve()


def test_resolve_user_workspace_path_accepts_bare_mount_root_alias(tmp_path: Path) -> None:
    from abstractruntime.integrations.abstractcore.workspace_scoped_tools import WorkspaceScope, resolve_user_workspace_path

    ws = tmp_path / "workspace"
    ws.mkdir(parents=True, exist_ok=True)
    mounted = tmp_path / "team-a" / "reports"
    mounted.mkdir(parents=True, exist_ok=True)

    mounts = build_workspace_mounts(allowed_dirs=[mounted], used_names=set())
    mount_alias = next(iter(mounts.keys()))

    scope = WorkspaceScope.from_input_data(
        {
            "workspace_root": str(ws),
            "workspace_access_mode": "workspace_or_allowed",
            "workspace_allowed_paths": [str(mounted)],
        }
    )
    assert scope is not None

    resolved = resolve_user_workspace_path(scope=scope, user_path=mount_alias)
    assert resolved.resolved_path == mounted.resolve()
    assert resolved.virtual_path == mount_alias
    assert resolved.root_path == mounted.resolve()


def test_all_except_ignored_absolute_paths_under_workspace_root_still_canonicalize(tmp_path: Path) -> None:
    from abstractruntime.integrations.abstractcore.workspace_scoped_tools import WorkspaceScope, resolve_user_workspace_path

    ws = tmp_path / "outside"
    ws.mkdir(parents=True, exist_ok=True)
    (ws / "secret.txt").write_text("secret\n", encoding="utf-8")

    scope = WorkspaceScope.from_input_data(
        {
            "workspace_root": str(ws),
            "workspace_access_mode": "all_except_ignored",
            "workspace_allowed_paths": [str(ws)],
        }
    )
    assert scope is not None

    resolved = resolve_user_workspace_path(scope=scope, user_path=str(ws / "secret.txt"))
    assert resolved.resolved_path == (ws / "secret.txt").resolve()
    assert resolved.virtual_path == "secret.txt"
    assert resolved.root_path == ws.resolve()



from __future__ import annotations

import pytest


def _tool_names(specs: list[object]) -> set[str]:
    out: set[str] = set()
    for s in specs:
        if not isinstance(s, dict):
            continue
        name = s.get("name")
        if isinstance(name, str) and name.strip():
            out.add(name.strip())
    return out


def test_comms_tools_are_opt_in(monkeypatch: pytest.MonkeyPatch) -> None:
    from abstractruntime.integrations.abstractcore.default_tools import list_default_tool_specs

    for k in (
        "ABSTRACT_ENABLE_COMMS_TOOLS",
        "ABSTRACT_ENABLE_EMAIL_TOOLS",
        "ABSTRACT_ENABLE_WHATSAPP_TOOLS",
        "ABSTRACT_ENABLE_TELEGRAM_TOOLS",
    ):
        monkeypatch.delenv(k, raising=False)

    specs = list_default_tool_specs()
    names = _tool_names(specs)
    assert "send_email" not in names
    assert "send_whatsapp_message" not in names
    assert "send_telegram_message" not in names

    monkeypatch.setenv("ABSTRACT_ENABLE_COMMS_TOOLS", "1")
    specs2 = list_default_tool_specs()
    names2 = _tool_names(specs2)
    assert "send_email" in names2
    assert "list_emails" in names2
    assert "read_email" in names2
    assert "send_whatsapp_message" in names2
    assert "list_whatsapp_messages" in names2
    assert "read_whatsapp_message" in names2
    assert "send_telegram_message" in names2
    assert "send_telegram_artifact" in names2



from __future__ import annotations

from pathlib import Path

from abstractcore.tools.common_tools import read_file
from abstractruntime.integrations.abstractcore.tool_executor import MappingToolExecutor


def test_mapping_tool_executor_maps_filename_to_file_path_for_read_file(tmp_path: Path) -> None:
    target = tmp_path / "demo.txt"
    target.write_text("hello\n", encoding="utf-8")

    exe = MappingToolExecutor.from_tools([read_file])
    result = exe.execute(
        tool_calls=[
            {
                "call_id": "1",
                "name": "read_file",
                "arguments": {"filename": str(target)},
            }
        ]
    )

    assert result["results"][0]["success"] is True
    out = result["results"][0]["output"]
    assert isinstance(out, str)
    assert "hello" in out






from __future__ import annotations

from typing import Any


def test_wait_event_can_carry_prompt_choices_and_allow_free_text() -> None:
    """Regression test: EVENT waits can be used for durable ask+wait UX.

    This is important for thin clients: the runtime must persist enough UI metadata
    (prompt/choices) for a remote host to render the question and resume later.
    """
    from abstractruntime.core.models import Effect, EffectType, RunStatus, StepPlan, WaitReason
    from abstractruntime.core.runtime import Runtime
    from abstractruntime.core.spec import WorkflowSpec
    from abstractruntime.storage.in_memory import InMemoryLedgerStore, InMemoryRunStore

    def node(run: Any, ctx: Any) -> StepPlan:
        del run, ctx
        return StepPlan(
            node_id="n1",
            effect=Effect(
                type=EffectType.WAIT_EVENT,
                payload={
                    "wait_key": "wk_ask",
                    "prompt": "Pick one:",
                    "choices": ["a", "b"],
                    "allow_free_text": False,
                },
                result_key="_temp.answer",
            ),
            next_node=None,
        )

    wf = WorkflowSpec(workflow_id="wf_wait_event_prompt", entry_node="n1", nodes={"n1": node})
    runtime = Runtime(run_store=InMemoryRunStore(), ledger_store=InMemoryLedgerStore())

    run_id = runtime.start(workflow=wf, vars={})
    state = runtime.tick(workflow=wf, run_id=run_id, max_steps=10)

    assert state.status == RunStatus.WAITING
    assert state.waiting is not None
    assert state.waiting.reason == WaitReason.EVENT
    assert state.waiting.wait_key == "wk_ask"
    assert state.waiting.prompt == "Pick one:"
    assert state.waiting.choices == ["a", "b"]
    assert state.waiting.allow_free_text is False

    # The durable ledger must include the same metadata for replay across reconnects.
    ledger = runtime.get_ledger(run_id)
    assert isinstance(ledger, list) and ledger
    last = ledger[-1]
    assert isinstance(last, dict)
    wait = (last.get("result") or {}).get("wait") if isinstance(last.get("result"), dict) else None
    assert isinstance(wait, dict)
    assert wait.get("reason") == "event"
    assert wait.get("wait_key") == "wk_ask"
    assert wait.get("prompt") == "Pick one:"
    assert wait.get("choices") == ["a", "b"]
    assert wait.get("allow_free_text") is False





"""Context node adapters (active context helpers).

These nodes mutate `run.vars["context"]` durably so pause/resume works and other
nodes (Agent/Subflow/LLM_CALL with use_context) can read the updated history.
"""

from __future__ import annotations

from typing import Any, Callable, Dict, Optional


def _persist_node_output(run_vars: Dict[str, Any], node_id: str, value: Dict[str, Any]) -> None:
    temp = run_vars.get("_temp")
    if not isinstance(temp, dict):
        temp = {}
        run_vars["_temp"] = temp
    persisted = temp.get("node_outputs")
    if not isinstance(persisted, dict):
        persisted = {}
        temp["node_outputs"] = persisted
    persisted[node_id] = value


def create_add_message_node_handler(
    *,
    node_id: str,
    next_node: Optional[str],
    data_aware_handler: Optional[Callable[[Any], Any]],
    flow: Any,
) -> Callable:
    """Create a handler for `add_message` visual nodes.

    Contract:
    - Inputs:
      - `role`: message role
      - `content`: message content
    - Behavior:
      - builds a canonical message object (timestamp + metadata.message_id)
      - appends it to `run.vars.context.messages` (durable)
      - does NOT clobber `_last_output` (pass-through)
    """
    from abstractruntime.core.models import StepPlan
    from ..compiler import _sync_effect_results_to_node_outputs

    def handler(run: Any, ctx: Any) -> "StepPlan":
        del ctx
        if flow is not None and hasattr(flow, "_node_outputs") and hasattr(flow, "_data_edge_map"):
            _sync_effect_results_to_node_outputs(run, flow)

        last_output = run.vars.get("_last_output", {})
        resolved = data_aware_handler(last_output) if callable(data_aware_handler) else {}
        payload = resolved if isinstance(resolved, dict) else {}

        message = payload.get("message")
        if not isinstance(message, dict):
            # Fallback: build a minimal message object if the visual handler did not.
            role = payload.get("role")
            role_str = role if isinstance(role, str) else str(role or "user")
            content = payload.get("content")
            content_str = content if isinstance(content, str) else str(content or "")

            try:
                from datetime import datetime, timezone

                timestamp = datetime.now(timezone.utc).isoformat()
            except Exception:
                from datetime import datetime

                timestamp = datetime.utcnow().isoformat() + "Z"

            import uuid

            message = {
                "role": role_str,
                "content": content_str,
                "timestamp": timestamp,
                "metadata": {"message_id": f"msg_{uuid.uuid4().hex}"},
            }

        ctx_ns = run.vars.get("context")
        if not isinstance(ctx_ns, dict):
            ctx_ns = {}
            run.vars["context"] = ctx_ns

        msgs = ctx_ns.get("messages")
        if not isinstance(msgs, list):
            msgs = []
            ctx_ns["messages"] = msgs

        msgs.append(message)

        # Persist node outputs for pause/resume and data-edge reads.
        _persist_node_output(
            run.vars,
            node_id,
            {
                "message": message,
                "context": ctx_ns,
                "task": ctx_ns.get("task") if isinstance(ctx_ns.get("task"), str) else str(ctx_ns.get("task") or ""),
                "messages": list(msgs),
            },
        )

        if next_node:
            return StepPlan(node_id=node_id, next_node=next_node)
        return StepPlan(node_id=node_id, complete_output={"success": True, "message": message})

    return handler




from __future__ import annotations

from pathlib import Path

import pytest

from abstractruntime.core.models import RunState, RunStatus, StepRecord, StepStatus
from abstractruntime.storage.commands import CommandRecord, InMemoryCommandStore, JsonlCommandStore
from abstractruntime.storage.in_memory import InMemoryLedgerStore, InMemoryRunStore
from abstractruntime.storage.json_files import JsonFileRunStore, JsonlLedgerStore
from abstractruntime.storage.ledger_chain import HashChainedLedgerStore
from abstractruntime.storage.observable import ObservableLedgerStore
from abstractruntime.storage.offloading import OffloadingLedgerStore, OffloadingRunStore
from abstractruntime.storage.artifacts import InMemoryArtifactStore
from abstractruntime.storage.sqlite import SqliteCommandStore, SqliteDatabase, SqliteLedgerStore, SqliteRunStore


def _run(run_id: str, *, parent_run_id: str | None = None) -> RunState:
    return RunState(
        run_id=run_id,
        workflow_id="wf",
        status=RunStatus.COMPLETED,
        current_node="done",
        vars={},
        parent_run_id=parent_run_id,
    )


def _record(run_id: str, step_id: str) -> StepRecord:
    return StepRecord(run_id=run_id, step_id=step_id, node_id="n", status=StepStatus.COMPLETED, result={"ok": True})


@pytest.mark.parametrize("kind", ["memory", "json", "sqlite"])
def test_run_store_delete_removes_checkpoint_index_and_children(kind: str, tmp_path: Path) -> None:
    if kind == "memory":
        store = InMemoryRunStore()
    elif kind == "json":
        store = JsonFileRunStore(tmp_path / "runs")
    else:
        store = SqliteRunStore(SqliteDatabase(tmp_path / "gateway.sqlite3"))

    root = _run("root")
    child = _run("child", parent_run_id="root")
    other = _run("other")
    store.save(root)
    store.save(child)
    store.save(other)

    assert store.load("child") is not None
    assert [r.run_id for r in store.list_children(parent_run_id="root")] == ["child"]

    assert store.delete("child") is True
    assert store.delete("child") is False
    assert store.load("child") is None
    assert store.list_children(parent_run_id="root") == []
    assert {r["run_id"] for r in store.list_run_index(limit=10)} == {"root", "other"}


@pytest.mark.parametrize("kind", ["memory", "json", "sqlite"])
def test_ledger_store_delete_removes_records_and_is_idempotent(kind: str, tmp_path: Path) -> None:
    if kind == "memory":
        store = InMemoryLedgerStore()
    elif kind == "json":
        store = JsonlLedgerStore(tmp_path / "ledger")
    else:
        store = SqliteLedgerStore(SqliteDatabase(tmp_path / "gateway.sqlite3"))

    store.append(_record("run_1", "s1"))
    store.append(_record("run_1", "s2"))
    store.append(_record("run_2", "s3"))

    assert len(store.list("run_1")) == 2
    assert store.delete("run_1") == 2
    assert store.delete("run_1") == 0
    assert store.list("run_1") == []
    assert len(store.list("run_2")) == 1


def test_wrapped_stores_delegate_delete_and_leave_artifacts_for_gateway_purge(tmp_path: Path) -> None:
    artifacts = InMemoryArtifactStore()
    run_store = OffloadingRunStore(JsonFileRunStore(tmp_path / "runs"), artifact_store=artifacts, max_inline_bytes=1)
    ledger_store = OffloadingLedgerStore(
        ObservableLedgerStore(HashChainedLedgerStore(JsonlLedgerStore(tmp_path / "ledger"))),
        artifact_store=artifacts,
        max_inline_bytes=1,
    )

    run = _run("run_1")
    run.vars["_temp"] = {"large": "x" * 64}
    run_store.save(run)
    ledger_store.append(_record("run_1", "s1"))

    assert artifacts.list_by_run("run_1")
    assert run_store.delete("run_1") is True
    assert ledger_store.delete("run_1") == 1
    assert artifacts.list_by_run("run_1")
    assert artifacts.delete_by_run("run_1") >= 1
    assert artifacts.list_by_run("run_1") == []


@pytest.mark.parametrize("kind", ["memory", "json", "sqlite"])
def test_command_store_delete_by_run_preserves_other_runs(kind: str, tmp_path: Path) -> None:
    if kind == "memory":
        store = InMemoryCommandStore()
    elif kind == "json":
        store = JsonlCommandStore(tmp_path / "commands")
    else:
        store = SqliteCommandStore(SqliteDatabase(tmp_path / "gateway.sqlite3"))

    store.append(CommandRecord(command_id="c1", run_id="run_1", type="pause", payload={}, ts="t"))
    store.append(CommandRecord(command_id="c2", run_id="run_1", type="resume", payload={}, ts="t"))
    store.append(CommandRecord(command_id="c3", run_id="run_2", type="cancel", payload={}, ts="t"))

    assert store.delete_by_run("run_1") == 2
    assert store.delete_by_run("run_1") == 0

    items, cursor = store.list_after(after=0, limit=10)
    assert [r.command_id for r in items] == ["c3"]
    assert cursor == items[-1].seq



"""abstractruntime.storage.snapshots

Named snapshots/bookmarks for durable run state.

A snapshot is a *user/operator-facing checkpoint* with:
- name/description/tags
- timestamps
- embedded (JSON) run state

This is intentionally simple for v0.1 (file-per-snapshot).
"""

from __future__ import annotations

import json
import uuid
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional

from ..core.models import RunState


def utc_now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


@dataclass
class Snapshot:
    snapshot_id: str
    run_id: str

    step_id: Optional[str] = None

    name: str = ""
    description: Optional[str] = None
    tags: List[str] = field(default_factory=list)

    created_at: str = field(default_factory=utc_now_iso)
    updated_at: str = field(default_factory=utc_now_iso)

    # JSON-only representation of the run state
    run_state: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_run(
        cls,
        *,
        run: RunState,
        name: str,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        step_id: Optional[str] = None,
    ) -> "Snapshot":
        return cls(
            snapshot_id=str(uuid.uuid4()),
            run_id=run.run_id,
            step_id=step_id,
            name=name,
            description=description,
            tags=list(tags or []),
            run_state=asdict(run),
        )


class SnapshotStore(ABC):
    @abstractmethod
    def save(self, snapshot: Snapshot) -> None: ...

    @abstractmethod
    def load(self, snapshot_id: str) -> Optional[Snapshot]: ...

    @abstractmethod
    def delete(self, snapshot_id: str) -> bool: ...

    @abstractmethod
    def list(
        self,
        *,
        run_id: Optional[str] = None,
        tag: Optional[str] = None,
        query: Optional[str] = None,
        limit: int = 100,
    ) -> List[Snapshot]: ...


class InMemorySnapshotStore(SnapshotStore):
    def __init__(self):
        self._snaps: Dict[str, Snapshot] = {}

    def save(self, snapshot: Snapshot) -> None:
        snapshot.updated_at = utc_now_iso()
        self._snaps[snapshot.snapshot_id] = snapshot

    def load(self, snapshot_id: str) -> Optional[Snapshot]:
        return self._snaps.get(snapshot_id)

    def delete(self, snapshot_id: str) -> bool:
        return self._snaps.pop(snapshot_id, None) is not None

    def list(
        self,
        *,
        run_id: Optional[str] = None,
        tag: Optional[str] = None,
        query: Optional[str] = None,
        limit: int = 100,
    ) -> List[Snapshot]:
        snaps = list(self._snaps.values())
        return _filter_snapshots(snaps, run_id=run_id, tag=tag, query=query, limit=limit)


class JsonSnapshotStore(SnapshotStore):
    """File-based snapshot store (one JSON file per snapshot)."""

    def __init__(self, base_dir: str | Path):
        self._base = Path(base_dir)
        self._base.mkdir(parents=True, exist_ok=True)

    def _path(self, snapshot_id: str) -> Path:
        return self._base / f"snapshot_{snapshot_id}.json"

    def save(self, snapshot: Snapshot) -> None:
        snapshot.updated_at = utc_now_iso()
        p = self._path(snapshot.snapshot_id)
        with p.open("w", encoding="utf-8") as f:
            json.dump(asdict(snapshot), f, ensure_ascii=False, indent=2)

    def load(self, snapshot_id: str) -> Optional[Snapshot]:
        p = self._path(snapshot_id)
        if not p.exists():
            return None
        with p.open("r", encoding="utf-8") as f:
            data = json.load(f)
        return Snapshot(
            snapshot_id=data["snapshot_id"],
            run_id=data["run_id"],
            step_id=data.get("step_id"),
            name=data.get("name") or "",
            description=data.get("description"),
            tags=list(data.get("tags") or []),
            created_at=data.get("created_at") or utc_now_iso(),
            updated_at=data.get("updated_at") or utc_now_iso(),
            run_state=dict(data.get("run_state") or {}),
        )

    def delete(self, snapshot_id: str) -> bool:
        p = self._path(snapshot_id)
        if not p.exists():
            return False
        p.unlink()
        return True

    def list(
        self,
        *,
        run_id: Optional[str] = None,
        tag: Optional[str] = None,
        query: Optional[str] = None,
        limit: int = 100,
    ) -> List[Snapshot]:
        snaps: List[Snapshot] = []
        for p in sorted(self._base.glob("snapshot_*.json")):
            try:
                with p.open("r", encoding="utf-8") as f:
                    data = json.load(f)
                snaps.append(
                    Snapshot(
                        snapshot_id=data["snapshot_id"],
                        run_id=data["run_id"],
                        step_id=data.get("step_id"),
                        name=data.get("name") or "",
                        description=data.get("description"),
                        tags=list(data.get("tags") or []),
                        created_at=data.get("created_at") or utc_now_iso(),
                        updated_at=data.get("updated_at") or utc_now_iso(),
                        run_state=dict(data.get("run_state") or {}),
                    )
                )
            except Exception:
                # Corrupt snapshot files are ignored for now (MVP); verification tools can be added later.
                continue

        return _filter_snapshots(snaps, run_id=run_id, tag=tag, query=query, limit=limit)


def _filter_snapshots(
    snaps: List[Snapshot],
    *,
    run_id: Optional[str],
    tag: Optional[str],
    query: Optional[str],
    limit: int,
) -> List[Snapshot]:
    out = snaps

    if run_id:
        out = [s for s in out if s.run_id == run_id]

    if tag:
        tag_l = tag.lower()
        out = [s for s in out if any(t.lower() == tag_l for t in (s.tags or []))]

    if query:
        q = query.lower()
        def _matches(s: Snapshot) -> bool:
            name = (s.name or "").lower()
            desc = (s.description or "").lower()
            return q in name or q in desc
        out = [s for s in out if _matches(s)]

    # Newest first (best-effort ISO ordering)
    out.sort(key=lambda s: s.created_at or "", reverse=True)
    return out[: max(0, int(limit))]




from __future__ import annotations

from pathlib import Path


def test_resolved_audio_artifact_preserves_content_type_for_transcription(tmp_path: Path) -> None:
    from abstractcore.providers.base import BaseProvider
    from abstractruntime.integrations.abstractcore.llm_client import _is_audio_media_item, _resolve_media_artifacts

    content_path = tmp_path / "artifact-content-without-extension"
    content_path.write_bytes(b"RIFF\x00\x00\x00\x00WAVE")

    class Store:
        def _content_path(self, artifact_id: str) -> Path:
            assert artifact_id == "audio-1"
            return content_path

    resolved = _resolve_media_artifacts(
        [{"$artifact": "audio-1", "content_type": "audio/wav"}],
        artifact_store=Store(),
    )

    assert isinstance(resolved, list)
    assert resolved == [
        {
            "$artifact": "audio-1",
            "artifact_id": "audio-1",
            "content_type": "audio/wav",
            "file_path": str(content_path),
            "mime_type": "audio/wav",
            "type": "audio",
        }
    ]
    assert _is_audio_media_item(resolved[0]) is True
    assert BaseProvider._media_type(resolved[0]) == "audio"


def test_effect_handler_materialized_audio_artifact_preserves_content_type(tmp_path: Path) -> None:
    from abstractcore.providers.base import BaseProvider
    from abstractruntime.integrations.abstractcore.effect_handlers import _resolve_llm_call_media

    class Artifact:
        content = b"RIFF\x00\x00\x00\x00WAVE"
        content_type = None
        metadata = None

    class Store:
        def load(self, artifact_id: str) -> Artifact:
            assert artifact_id == "audio-1"
            return Artifact()

    resolved, error = _resolve_llm_call_media(
        [{"$artifact": "audio-1", "content_type": "audio/wav"}],
        artifact_store=Store(),
        temp_dir=tmp_path,
    )

    assert error is None
    assert isinstance(resolved, list)
    assert len(resolved) == 1
    item = resolved[0]
    assert item["$artifact"] == "audio-1"
    assert item["artifact_id"] == "audio-1"
    assert item["content_type"] == "audio/wav"
    assert item["type"] == "audio"
    assert Path(item["file_path"]).suffix == ".wav"
    assert BaseProvider._media_type(item) == "audio"


def test_effect_handler_materialized_image_artifacts_preserve_roles(tmp_path: Path) -> None:
    from abstractruntime.integrations.abstractcore.effect_handlers import _resolve_llm_call_media

    class Artifact:
        content = PNG_1X1 = (
            b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR"
            b"\x00\x00\x00\x01\x00\x00\x00\x01\x08\x06\x00\x00\x00"
            b"\x1f\x15\xc4\x89"
        )
        content_type = "image/png"
        metadata = None

    class Store:
        def load(self, artifact_id: str) -> Artifact:
            assert artifact_id in {"source-img", "mask-img"}
            return Artifact()

    resolved, error = _resolve_llm_call_media(
        [
            {"$artifact": "source-img", "content_type": "image/png", "role": "source"},
            {"$artifact": "mask-img", "content_type": "image/png", "role": "mask"},
        ],
        artifact_store=Store(),
        temp_dir=tmp_path,
    )

    assert error is None
    assert isinstance(resolved, list)
    assert [item.get("role") for item in resolved] == ["source", "mask"]
    assert all(item.get("type") == "image" for item in resolved)
    assert all(Path(str(item.get("file_path") or "")).is_file() for item in resolved)



from __future__ import annotations

import re
from pathlib import Path


def _extract_optional_dependency_block(text: str, *, key: str) -> str:
    lines = text.splitlines()
    start = None
    for i, line in enumerate(lines):
        if line.startswith(f"{key} = ["):
            start = i
            break
    assert start is not None, f"Missing optional-dependencies entry: {key}"

    block: list[str] = []
    for line in lines[start + 1 :]:
        if line.strip() == "]":
            break
        block.append(line)
    return "\n".join(block)


def _extract_optional_dependency_keys(text: str) -> set[str]:
    in_section = False
    keys: set[str] = set()
    for line in text.splitlines():
        stripped = line.strip()
        if stripped == "[project.optional-dependencies]":
            in_section = True
            continue
        if in_section and stripped.startswith("[") and stripped.endswith("]"):
            break
        if not in_section:
            continue
        match = re.match(r"^([A-Za-z0-9_-]+)\s*=\s*\[", line)
        if match:
            keys.add(match.group(1))
    return keys


def test_runtime_base_is_remote_light_with_multimodal_and_mcp_support() -> None:
    pyproject = Path(__file__).resolve().parents[1] / "pyproject.toml"
    text = pyproject.read_text(encoding="utf-8")

    assert '"abstractsemantics>=0.0.3"' in text
    assert '"AbstractMemory>=0.2.6"' in text
    assert '"AbstractMemory[lancedb]' not in text
    assert "[project.optional-dependencies]" in text
    assert "abstractcore = [" not in text
    assert "multimodal = [" not in text
    assert "mcp-worker = [" not in text
    assert '"abstractcore[remote,tools,vision,voice,audio,music]>=2.13.38"' in text
    assert '"openai<2.0.0,>=1.109.1"' in text
    assert '"httpx<1.0.0,>=0.28.1"' in text
    assert '"anyio<5.0.0,>=4.12.1"' in text
    assert '"Pillow<13.0.0,>=10.0.0"' in text
    assert '"pypdf<7.0.0,>=6.0.0"' in text
    assert '"reportlab<5.0.0,>=4.0.0"' in text
    assert '"unstructured[docx,odt,pptx,rtf,xlsx]<0.19.0,>=0.18.32"' in text
    assert '"python-pptx<2.0.0,>=1.0.2"' in text
    assert "pymupdf" not in text.lower()
    assert "pymupdf4llm" not in text
    assert "pymupdf-layout" not in text
    assert '"torch' not in text
    assert '"sentence-transformers' not in text
    assert '"vllm' not in text
    assert '"mlx' not in text


def test_runtime_exposes_only_apple_and_gpu_user_install_profiles() -> None:
    pyproject = Path(__file__).resolve().parents[1] / "pyproject.toml"
    text = pyproject.read_text(encoding="utf-8")

    optional_keys = _extract_optional_dependency_keys(text)

    assert "apple" in optional_keys
    assert "gpu" in optional_keys
    assert "all-apple" not in optional_keys
    assert "all-gpu" not in optional_keys
    assert "abstractcore" not in optional_keys
    assert "multimodal" not in optional_keys
    assert "mcp-worker" not in optional_keys

    apple_block = _extract_optional_dependency_block(text, key="apple")
    gpu_block = _extract_optional_dependency_block(text, key="gpu")

    assert '"abstractcore[all-apple]>=2.13.38"' in apple_block
    assert '"abstractcore[all-gpu]>=2.13.38"' in gpu_block
    assert '"setuptools<82.0.0,>=80.10.2"' in apple_block
    assert '"setuptools<82.0.0,>=80.10.2"' in gpu_block
    assert "pymupdf" not in apple_block.lower()
    assert "pymupdf" not in gpu_block.lower()

    core_pyproject = pyproject.resolve().parents[1] / "abstractcore" / "pyproject.toml"
    if core_pyproject.exists():
        core_text = core_pyproject.read_text(encoding="utf-8")
        for core_profile in ("all-apple", "all-gpu"):
            core_block = _extract_optional_dependency_block(core_text, key=core_profile)
            assert '"pypdf>=6.0.0,<7.0.0"' in core_block
            assert "pymupdf" not in core_block.lower()



"""abstractruntime.storage.sqlite

SQLite-backed durability stores for production-oriented single-host deployments.

Design goals:
- Keep the durable execution substrate dependency-light (stdlib `sqlite3`).
- Provide restart-safe storage with real indexing (avoid directory scans + JSON parsing loops).
- Preserve the existing store interfaces (RunStore/LedgerStore/CommandStore) so hosts can
  switch backends without rewriting runtime logic.

Scope (backlog 446):
- RunStore (checkpointed RunState JSON)
- LedgerStore (append-only StepRecord JSON with per-run seq)
- CommandStore + CommandCursorStore (durable inbox + consumer cursor)
- WAIT_UNTIL index (wait_index table) so runners can query due runs efficiently.
"""

from __future__ import annotations

import json
import logging
import sqlite3
import threading
import uuid
from dataclasses import asdict
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from ..core.models import RunState, RunStatus, StepRecord, WaitReason, WaitState
from ..core.run_lifecycle import run_lifecycle_index_fields
from .base import LedgerStore, RunStore
from .commands import CommandAppendResult, CommandCursorStore, CommandRecord, CommandStore

logger = logging.getLogger(__name__)


def _utc_now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def _is_json_value(value: Any) -> bool:
    if value is None or isinstance(value, (str, int, float, bool)):
        return True
    if isinstance(value, list):
        return all(_is_json_value(v) for v in value)
    if isinstance(value, dict):
        return all(isinstance(k, str) and _is_json_value(v) for k, v in value.items())
    return False


def _runstate_from_dict(data: Dict[str, Any]) -> RunState:
    raw_status = data.get("status")
    status = raw_status if isinstance(raw_status, RunStatus) else RunStatus(str(raw_status))

    waiting: Optional[WaitState] = None
    raw_waiting = data.get("waiting")
    if isinstance(raw_waiting, dict):
        raw_reason = raw_waiting.get("reason")
        if raw_reason is None:
            raise ValueError("Persisted waiting state missing 'reason'")
        reason = raw_reason if isinstance(raw_reason, WaitReason) else WaitReason(str(raw_reason))
        waiting = WaitState(
            reason=reason,
            wait_key=raw_waiting.get("wait_key"),
            until=raw_waiting.get("until"),
            resume_to_node=raw_waiting.get("resume_to_node"),
            result_key=raw_waiting.get("result_key"),
            prompt=raw_waiting.get("prompt"),
            choices=raw_waiting.get("choices"),
            allow_free_text=bool(raw_waiting.get("allow_free_text", True)),
            details=raw_waiting.get("details"),
        )

    return RunState(
        run_id=str(data.get("run_id") or ""),
        workflow_id=str(data.get("workflow_id") or ""),
        status=status,
        current_node=str(data.get("current_node") or ""),
        vars=dict(data.get("vars") or {}),
        waiting=waiting,
        output=data.get("output"),
        error=data.get("error"),
        created_at=str(data.get("created_at") or ""),
        updated_at=str(data.get("updated_at") or ""),
        actor_id=data.get("actor_id"),
        session_id=data.get("session_id"),
        parent_run_id=data.get("parent_run_id"),
    )


class SqliteDatabase:
    """Small helper around a SQLite file with per-thread connections."""

    def __init__(self, path: str | Path) -> None:
        self._path = Path(path).expanduser().resolve()
        self._path.parent.mkdir(parents=True, exist_ok=True)
        self._local = threading.local()
        self._init_lock = threading.Lock()
        self._initialized = False
        self._ensure_schema()

    @property
    def path(self) -> Path:
        return self._path

    def connection(self) -> sqlite3.Connection:
        conn = getattr(self._local, "conn", None)
        if conn is None:
            conn = sqlite3.connect(str(self._path), timeout=30.0)
            conn.row_factory = sqlite3.Row
            self._apply_pragmas(conn)
            self._local.conn = conn
        return conn

    def _apply_pragmas(self, conn: sqlite3.Connection) -> None:
        # WAL improves writer/reader concurrency for the API+runner split.
        try:
            conn.execute("PRAGMA journal_mode=WAL;")
        except Exception:
            pass
        try:
            conn.execute("PRAGMA synchronous=NORMAL;")
        except Exception:
            pass
        try:
            conn.execute("PRAGMA foreign_keys=ON;")
        except Exception:
            pass
        try:
            conn.execute("PRAGMA busy_timeout=5000;")
        except Exception:
            pass

    def _ensure_schema(self) -> None:
        with self._init_lock:
            if self._initialized:
                return
            conn = sqlite3.connect(str(self._path), timeout=30.0)
            try:
                conn.row_factory = sqlite3.Row
                self._apply_pragmas(conn)

                # --- Runs ---
                conn.execute(
                    """
                    CREATE TABLE IF NOT EXISTS runs (
                      run_id TEXT PRIMARY KEY,
                      workflow_id TEXT NOT NULL,
                      status TEXT NOT NULL,
                      wait_reason TEXT,
                      wait_until TEXT,
                      parent_run_id TEXT,
                      actor_id TEXT,
                      session_id TEXT,
                      created_at TEXT,
                      updated_at TEXT,
                      run_json TEXT NOT NULL
                    );
                    """
                )
                conn.execute("CREATE INDEX IF NOT EXISTS idx_runs_status_updated ON runs(status, updated_at DESC);")
                conn.execute("CREATE INDEX IF NOT EXISTS idx_runs_workflow_updated ON runs(workflow_id, updated_at DESC);")
                conn.execute("CREATE INDEX IF NOT EXISTS idx_runs_parent ON runs(parent_run_id);")
                conn.execute("CREATE INDEX IF NOT EXISTS idx_runs_waiting ON runs(status, wait_reason, wait_until);")

                # --- WAIT_UNTIL index (scheduler) ---
                conn.execute(
                    """
                    CREATE TABLE IF NOT EXISTS wait_index (
                      run_id TEXT PRIMARY KEY,
                      next_due_iso TEXT NOT NULL,
                      updated_at_iso TEXT NOT NULL,
                      status TEXT
                    );
                    """
                )
                conn.execute("CREATE INDEX IF NOT EXISTS idx_wait_index_next_due ON wait_index(next_due_iso);")

                # --- Ledger ---
                conn.execute(
                    """
                    CREATE TABLE IF NOT EXISTS ledger (
                      run_id TEXT NOT NULL,
                      seq INTEGER NOT NULL,
                      record_json TEXT NOT NULL,
                      PRIMARY KEY (run_id, seq)
                    );
                    """
                )
                conn.execute("CREATE INDEX IF NOT EXISTS idx_ledger_run_seq ON ledger(run_id, seq);")
                conn.execute(
                    """
                    CREATE TABLE IF NOT EXISTS ledger_heads (
                      run_id TEXT PRIMARY KEY,
                      last_seq INTEGER NOT NULL
                    );
                    """
                )
                # Backfill ledger_heads for upgraded DBs.
                #
                # When `ledger_heads` is introduced after `ledger` already has rows, future appends must
                # continue from the existing MAX(seq) for each run_id (otherwise we would re-allocate
                # seq starting at 1 and hit `UNIQUE constraint failed: ledger.run_id, ledger.seq`).
                #
                # This is idempotent and safe to run on every startup.
                conn.execute(
                    """
                    INSERT INTO ledger_heads (run_id, last_seq)
                    SELECT run_id, MAX(seq) AS last_seq
                    FROM ledger
                    GROUP BY run_id
                    ON CONFLICT(run_id) DO UPDATE SET last_seq =
                      CASE
                        WHEN excluded.last_seq > ledger_heads.last_seq THEN excluded.last_seq
                        ELSE ledger_heads.last_seq
                      END;
                    """
                )

                # --- Commands (durable inbox) ---
                conn.execute(
                    """
                    CREATE TABLE IF NOT EXISTS commands (
                      seq INTEGER PRIMARY KEY AUTOINCREMENT,
                      command_id TEXT NOT NULL UNIQUE,
                      run_id TEXT NOT NULL,
                      type TEXT NOT NULL,
                      payload_json TEXT NOT NULL,
                      ts TEXT NOT NULL,
                      client_id TEXT
                    );
                    """
                )
                conn.execute("CREATE INDEX IF NOT EXISTS idx_commands_run_id ON commands(run_id);")

                # --- Command consumer cursors ---
                conn.execute(
                    """
                    CREATE TABLE IF NOT EXISTS command_cursors (
                      consumer_id TEXT PRIMARY KEY,
                      cursor INTEGER NOT NULL,
                      updated_at TEXT NOT NULL
                    );
                    """
                )

                conn.commit()
                self._initialized = True
            finally:
                try:
                    conn.close()
                except Exception:
                    pass


class SqliteRunStore(RunStore):
    """SQLite-backed RunStore with QueryableRunStore methods and a WAIT_UNTIL index."""

    def __init__(self, db: SqliteDatabase) -> None:
        self._db = db

    def save(self, run: RunState) -> None:
        wait_reason: Optional[str] = None
        wait_until: Optional[str] = None
        if run.waiting is not None:
            try:
                wait_reason = str(getattr(run.waiting.reason, "value", run.waiting.reason))
            except Exception:
                wait_reason = None
            if run.waiting.reason == WaitReason.UNTIL:
                wait_until = str(run.waiting.until) if run.waiting.until else None

        payload = json.dumps(asdict(run), ensure_ascii=False)

        conn = self._db.connection()
        with conn:
            conn.execute(
                """
                INSERT INTO runs (
                  run_id, workflow_id, status, wait_reason, wait_until,
                  parent_run_id, actor_id, session_id,
                  created_at, updated_at,
                  run_json
                )
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                ON CONFLICT(run_id) DO UPDATE SET
                  workflow_id=excluded.workflow_id,
                  status=excluded.status,
                  wait_reason=excluded.wait_reason,
                  wait_until=excluded.wait_until,
                  parent_run_id=excluded.parent_run_id,
                  actor_id=excluded.actor_id,
                  session_id=excluded.session_id,
                  updated_at=excluded.updated_at,
                  run_json=excluded.run_json;
                """,
                (
                    str(run.run_id),
                    str(run.workflow_id),
                    str(getattr(run.status, "value", run.status)),
                    wait_reason,
                    wait_until,
                    str(run.parent_run_id) if run.parent_run_id else None,
                    str(run.actor_id) if run.actor_id else None,
                    str(run.session_id) if run.session_id else None,
                    str(run.created_at),
                    str(run.updated_at),
                    payload,
                ),
            )

            # Maintain WAIT_UNTIL index (only applies to WAITING runs).
            if run.status == RunStatus.WAITING and wait_reason == WaitReason.UNTIL.value and wait_until:
                conn.execute(
                    """
                    INSERT INTO wait_index (run_id, next_due_iso, updated_at_iso, status)
                    VALUES (?, ?, ?, ?)
                    ON CONFLICT(run_id) DO UPDATE SET
                      next_due_iso=excluded.next_due_iso,
                      updated_at_iso=excluded.updated_at_iso,
                      status=excluded.status;
                    """,
                    (str(run.run_id), str(wait_until), str(run.updated_at), "waiting_until"),
                )
            else:
                conn.execute("DELETE FROM wait_index WHERE run_id = ?;", (str(run.run_id),))

    def load(self, run_id: str) -> Optional[RunState]:
        rid = str(run_id or "").strip()
        if not rid:
            return None
        conn = self._db.connection()
        row = conn.execute("SELECT run_json FROM runs WHERE run_id = ?;", (rid,)).fetchone()
        if row is None:
            return None
        try:
            data = json.loads(str(row["run_json"] or "{}"))
        except Exception:
            return None
        if not isinstance(data, dict):
            return None
        try:
            return _runstate_from_dict(data)
        except Exception:
            return None

    def delete(self, run_id: str) -> bool:
        rid = str(run_id or "").strip()
        if not rid:
            return False
        conn = self._db.connection()
        with conn:
            cur = conn.execute("DELETE FROM runs WHERE run_id = ?;", (rid,))
            conn.execute("DELETE FROM wait_index WHERE run_id = ?;", (rid,))
        return int(getattr(cur, "rowcount", 0) or 0) > 0

    # --- QueryableRunStore methods ---

    def list_runs(
        self,
        *,
        status: Optional[RunStatus] = None,
        wait_reason: Optional[WaitReason] = None,
        workflow_id: Optional[str] = None,
        limit: int = 100,
    ) -> List[RunState]:
        clauses: list[str] = []
        params: list[Any] = []

        if status is not None:
            clauses.append("status = ?")
            params.append(str(getattr(status, "value", status)))
        if workflow_id is not None:
            clauses.append("workflow_id = ?")
            params.append(str(workflow_id))
        if wait_reason is not None:
            clauses.append("wait_reason = ?")
            params.append(str(getattr(wait_reason, "value", wait_reason)))

        where = "WHERE " + " AND ".join(clauses) if clauses else ""
        lim = max(1, int(limit or 100))

        conn = self._db.connection()
        rows = conn.execute(
            f"SELECT run_json FROM runs {where} ORDER BY updated_at DESC LIMIT ?;",
            (*params, lim),
        ).fetchall()

        out: List[RunState] = []
        for row in rows or []:
            try:
                data = json.loads(str(row["run_json"] or "{}"))
            except Exception:
                continue
            if not isinstance(data, dict):
                continue
            try:
                out.append(_runstate_from_dict(data))
            except Exception:
                continue
        return out

    def list_run_index(
        self,
        *,
        status: Optional[RunStatus] = None,
        workflow_id: Optional[str] = None,
        session_id: Optional[str] = None,
        root_only: bool = False,
        limit: int = 100,
    ) -> List[Dict[str, Any]]:
        clauses: list[str] = []
        params: list[Any] = []

        if status is not None:
            clauses.append("status = ?")
            params.append(str(getattr(status, "value", status)))
        if workflow_id is not None:
            clauses.append("workflow_id = ?")
            params.append(str(workflow_id))
        if session_id is not None:
            clauses.append("session_id = ?")
            params.append(str(session_id))
        if bool(root_only):
            clauses.append("(parent_run_id IS NULL OR parent_run_id = '')")

        where = "WHERE " + " AND ".join(clauses) if clauses else ""
        lim = max(1, int(limit or 100))

        conn = self._db.connection()
        rows = conn.execute(
            f"""
            SELECT
              run_id, workflow_id, status,
              wait_reason, wait_until,
              parent_run_id, actor_id, session_id,
              created_at, updated_at,
              run_json
            FROM runs
            {where}
            ORDER BY updated_at DESC
            LIMIT ?;
            """,
            (*params, lim),
        ).fetchall()

        out: List[Dict[str, Any]] = []
        for row in rows or []:
            try:
                run_payload = json.loads(str(row["run_json"] or "{}"))
            except Exception:
                run_payload = {}
            vars_obj = run_payload.get("vars") if isinstance(run_payload, dict) else None
            out.append(
                {
                    "run_id": str(row["run_id"] or ""),
                    "workflow_id": str(row["workflow_id"] or ""),
                    "status": str(row["status"] or ""),
                    "wait_reason": str(row["wait_reason"] or "") or None,
                    "wait_until": str(row["wait_until"] or "") or None,
                    "parent_run_id": str(row["parent_run_id"] or "") or None,
                    "actor_id": str(row["actor_id"] or "") or None,
                    "session_id": str(row["session_id"] or "") or None,
                    "created_at": str(row["created_at"] or "") or None,
                    "updated_at": str(row["updated_at"] or "") or None,
                    **run_lifecycle_index_fields(vars_obj),
                }
            )
        return out

    def list_due_wait_until(
        self,
        *,
        now_iso: str,
        limit: int = 100,
    ) -> List[RunState]:
        now = str(now_iso or "").strip()
        lim = max(1, int(limit or 100))
        conn = self._db.connection()
        rows = conn.execute(
            """
            SELECT r.run_json
            FROM wait_index w
            JOIN runs r ON r.run_id = w.run_id
            WHERE w.next_due_iso <= ?
            ORDER BY w.next_due_iso ASC
            LIMIT ?;
            """,
            (now, lim),
        ).fetchall()

        out: List[RunState] = []
        for row in rows or []:
            try:
                data = json.loads(str(row["run_json"] or "{}"))
            except Exception:
                continue
            if not isinstance(data, dict):
                continue
            try:
                out.append(_runstate_from_dict(data))
            except Exception:
                continue
        return out

    def list_children(
        self,
        *,
        parent_run_id: str,
        status: Optional[RunStatus] = None,
    ) -> List[RunState]:
        parent = str(parent_run_id or "").strip()
        if not parent:
            return []
        clauses = ["parent_run_id = ?"]
        params: list[Any] = [parent]

        if status is not None:
            clauses.append("status = ?")
            params.append(str(getattr(status, "value", status)))

        where = "WHERE " + " AND ".join(clauses)
        conn = self._db.connection()
        rows = conn.execute(
            f"SELECT run_json FROM runs {where} ORDER BY created_at ASC, run_id ASC;",
            tuple(params),
        ).fetchall()

        out: List[RunState] = []
        for row in rows or []:
            try:
                data = json.loads(str(row["run_json"] or "{}"))
            except Exception:
                continue
            if not isinstance(data, dict):
                continue
            try:
                out.append(_runstate_from_dict(data))
            except Exception:
                continue
        return out


class SqliteLedgerStore(LedgerStore):
    """SQLite-backed append-only ledger store with per-run seq."""

    def __init__(self, db: SqliteDatabase) -> None:
        self._db = db

    def append(self, record: StepRecord) -> None:
        run_id = str(record.run_id or "").strip()
        if not run_id:
            raise ValueError("StepRecord.run_id must be non-empty")

        payload = json.dumps(asdict(record), ensure_ascii=False)
        conn = self._db.connection()
        with conn:
            conn.execute(
                """
                INSERT INTO ledger_heads (run_id, last_seq)
                VALUES (?, 0)
                ON CONFLICT(run_id) DO NOTHING;
                """,
                (run_id,),
            )
            conn.execute("UPDATE ledger_heads SET last_seq = last_seq + 1 WHERE run_id = ?;", (run_id,))
            row = conn.execute("SELECT last_seq FROM ledger_heads WHERE run_id = ?;", (run_id,)).fetchone()
            if row is None:
                raise RuntimeError("Failed to allocate ledger seq")
            seq = int(row["last_seq"] or 0)
            try:
                conn.execute(
                    "INSERT INTO ledger (run_id, seq, record_json) VALUES (?, ?, ?);",
                    (run_id, int(seq), payload),
                )
            except sqlite3.IntegrityError as e:
                msg = str(e)
                if "UNIQUE constraint failed: ledger.run_id, ledger.seq" in msg:
                    logger.error(
                        "SqliteLedgerStore.append failed due to duplicate seq allocation "
                        "(run_id=%s seq=%s db=%s). This usually indicates concurrent writers running "
                        "a non-atomic seq allocator or a corrupted ledger_heads table.",
                        run_id,
                        seq,
                        self._db.path,
                    )
                raise sqlite3.IntegrityError(f"{msg} (run_id={run_id!r}, seq={seq}, db={self._db.path})") from e

    def list(self, run_id: str) -> List[Dict[str, Any]]:
        rid = str(run_id or "").strip()
        if not rid:
            return []
        conn = self._db.connection()
        rows = conn.execute(
            "SELECT record_json FROM ledger WHERE run_id = ? ORDER BY seq ASC;",
            (rid,),
        ).fetchall()
        out: List[Dict[str, Any]] = []
        for row in rows or []:
            try:
                obj = json.loads(str(row["record_json"] or "{}"))
            except Exception:
                continue
            if isinstance(obj, dict):
                out.append(obj)
        return out

    def delete(self, run_id: str) -> int:
        rid = str(run_id or "").strip()
        if not rid:
            return 0
        conn = self._db.connection()
        with conn:
            cur = conn.execute("DELETE FROM ledger WHERE run_id = ?;", (rid,))
            conn.execute("DELETE FROM ledger_heads WHERE run_id = ?;", (rid,))
        return int(getattr(cur, "rowcount", 0) or 0)

    def count(self, run_id: str) -> int:
        rid = str(run_id or "").strip()
        if not rid:
            return 0
        conn = self._db.connection()
        row = conn.execute("SELECT last_seq FROM ledger_heads WHERE run_id = ?;", (rid,)).fetchone()
        if row is None:
            return 0
        try:
            return int(row["last_seq"] or 0)
        except Exception:
            return 0

    def count_many(self, run_ids: List[str]) -> Dict[str, int]:
        ids = [str(r or "").strip() for r in (run_ids or []) if str(r or "").strip()]
        if not ids:
            return {}
        # SQLite parameter limit is high enough for typical UI pages; chunk defensively.
        out: Dict[str, int] = {}
        conn = self._db.connection()
        for i in range(0, len(ids), 900):
            chunk = ids[i : i + 900]
            q = ",".join(["?"] * len(chunk))
            rows = conn.execute(f"SELECT run_id, last_seq FROM ledger_heads WHERE run_id IN ({q});", tuple(chunk)).fetchall()
            for row in rows or []:
                rid = str(row["run_id"] or "").strip()
                if not rid:
                    continue
                try:
                    out[rid] = int(row["last_seq"] or 0)
                except Exception:
                    out[rid] = 0
        return out

    def metrics_many(self, run_ids: List[str]) -> Dict[str, Dict[str, int]]:
        """Return best-effort per-run metrics derived from completed ledger records."""
        ids = [str(r or "").strip() for r in (run_ids or []) if str(r or "").strip()]
        if not ids:
            return {}
        out: Dict[str, Dict[str, int]] = {}
        conn = self._db.connection()
        for i in range(0, len(ids), 300):
            chunk = ids[i : i + 300]
            q = ",".join(["?"] * len(chunk))
            rows = conn.execute(
                f"""
                WITH completed AS (
                  SELECT run_id, record_json
                  FROM ledger
                  WHERE run_id IN ({q})
                    AND json_extract(record_json, '$.status') = 'completed'
                )
                SELECT
                  run_id AS run_id,
                  COUNT(*) AS steps,
                  SUM(CASE WHEN json_extract(record_json, '$.effect.type') = 'llm_call' THEN 1 ELSE 0 END) AS llm_calls,
                  SUM(
                    CASE
                      WHEN json_extract(record_json, '$.effect.type') = 'tool_calls'
                        THEN COALESCE(json_array_length(json_extract(record_json, '$.effect.payload.tool_calls')), 0)
                      ELSE 0
                    END
                  ) AS tool_calls,
                  SUM(
                    CASE
                      WHEN json_extract(record_json, '$.effect.type') = 'llm_call'
                        THEN COALESCE(json_extract(record_json, '$.result.usage.total_tokens'), 0)
                      ELSE 0
                    END
                  ) AS tokens_total
                FROM completed
                GROUP BY run_id;
                """,
                tuple(chunk),
            ).fetchall()
            for row in rows or []:
                rid = str(row["run_id"] or "").strip()
                if not rid:
                    continue
                def _i(v: Any) -> int:
                    try:
                        return int(v or 0)
                    except Exception:
                        return 0
                out[rid] = {
                    "steps": _i(row["steps"]),
                    "llm_calls": _i(row["llm_calls"]),
                    "tool_calls": _i(row["tool_calls"]),
                    "tokens_total": _i(row["tokens_total"]),
                }
        return out

    def list_after(self, *, run_id: str, after: int, limit: int = 1000) -> Tuple[List[Dict[str, Any]], int]:
        """Optional cursor API (not part of LedgerStore ABC).

        Cursor semantics match the existing gateway API: `after` is the last consumed seq.
        """
        rid = str(run_id or "").strip()
        a = int(after or 0)
        lim = max(1, int(limit or 1000))
        if not rid:
            return ([], a)
        conn = self._db.connection()
        rows = conn.execute(
            "SELECT seq, record_json FROM ledger WHERE run_id = ? AND seq > ? ORDER BY seq ASC LIMIT ?;",
            (rid, a, lim),
        ).fetchall()
        out: List[Dict[str, Any]] = []
        next_cursor = a
        for row in rows or []:
            try:
                obj = json.loads(str(row["record_json"] or "{}"))
            except Exception:
                continue
            if not isinstance(obj, dict):
                continue
            out.append(obj)
            try:
                next_cursor = max(next_cursor, int(row["seq"] or next_cursor))
            except Exception:
                pass
        return (out, next_cursor)


class SqliteCommandStore(CommandStore):
    """SQLite-backed CommandStore (append-only, idempotent by command_id)."""

    def __init__(self, db: SqliteDatabase) -> None:
        self._db = db

    def append(self, record: CommandRecord) -> CommandAppendResult:
        cid = str(record.command_id or "").strip() or uuid.uuid4().hex
        run_id = str(record.run_id or "").strip()
        typ = str(record.type or "").strip()
        payload = dict(record.payload or {})
        ts = str(record.ts or "").strip() or _utc_now_iso()
        client_id = str(record.client_id).strip() if isinstance(record.client_id, str) and record.client_id else None

        if not run_id:
            raise ValueError("CommandRecord.run_id must be non-empty")
        if not typ:
            raise ValueError("CommandRecord.type must be non-empty")
        if not isinstance(payload, dict) or not _is_json_value(payload):
            raise ValueError("CommandRecord.payload must be a JSON-serializable dict")

        conn = self._db.connection()
        with conn:
            cur = conn.execute(
                """
                INSERT OR IGNORE INTO commands (command_id, run_id, type, payload_json, ts, client_id)
                VALUES (?, ?, ?, ?, ?, ?);
                """,
                (cid, run_id, typ, json.dumps(payload, ensure_ascii=False), ts, client_id),
            )
            if int(cur.rowcount or 0) == 1:
                seq = int(cur.lastrowid or 0)
                return CommandAppendResult(accepted=True, duplicate=False, seq=seq)

            row = conn.execute("SELECT seq FROM commands WHERE command_id = ?;", (cid,)).fetchone()
            seq2 = int(row["seq"]) if row is not None else 0
            return CommandAppendResult(accepted=False, duplicate=True, seq=seq2)

    def list_after(self, *, after: int, limit: int = 1000) -> Tuple[List[CommandRecord], int]:
        a = int(after or 0)
        lim = max(1, int(limit or 1000))
        conn = self._db.connection()
        rows = conn.execute(
            """
            SELECT seq, command_id, run_id, type, payload_json, ts, client_id
            FROM commands
            WHERE seq > ?
            ORDER BY seq ASC
            LIMIT ?;
            """,
            (a, lim),
        ).fetchall()

        out: List[CommandRecord] = []
        next_cursor = a
        for row in rows or []:
            try:
                payload = json.loads(str(row["payload_json"] or "{}"))
            except Exception:
                payload = {}
            if not isinstance(payload, dict):
                payload = {}
            seq = int(row["seq"] or 0)
            out.append(
                CommandRecord(
                    command_id=str(row["command_id"] or ""),
                    run_id=str(row["run_id"] or ""),
                    type=str(row["type"] or ""),
                    payload=payload,
                    ts=str(row["ts"] or ""),
                    client_id=str(row["client_id"] or "") or None,
                    seq=seq,
                )
            )
            next_cursor = max(next_cursor, seq)
        return (out, next_cursor)

    def get_last_seq(self) -> int:
        conn = self._db.connection()
        row = conn.execute("SELECT COALESCE(MAX(seq), 0) AS max_seq FROM commands;").fetchone()
        if row is None:
            return 0
        try:
            return int(row["max_seq"] or 0)
        except Exception:
            return 0

    def delete_by_run(self, run_id: str) -> int:
        rid = str(run_id or "").strip()
        if not rid:
            return 0
        conn = self._db.connection()
        with conn:
            cur = conn.execute("DELETE FROM commands WHERE run_id = ?;", (rid,))
        return int(getattr(cur, "rowcount", 0) or 0)


class SqliteCommandCursorStore(CommandCursorStore):
    """SQLite-backed durable cursor store for CommandStore replay."""

    def __init__(self, db: SqliteDatabase, *, consumer_id: str = "gateway_runner") -> None:
        self._db = db
        self._consumer_id = str(consumer_id or "gateway_runner").strip() or "gateway_runner"

    def load(self) -> int:
        conn = self._db.connection()
        row = conn.execute(
            "SELECT cursor FROM command_cursors WHERE consumer_id = ?;",
            (self._consumer_id,),
        ).fetchone()
        if row is None:
            return 0
        try:
            return int(row["cursor"] or 0)
        except Exception:
            return 0

    def save(self, cursor: int) -> None:
        cur = int(cursor or 0)
        conn = self._db.connection()
        with conn:
            conn.execute(
                """
                INSERT INTO command_cursors (consumer_id, cursor, updated_at)
                VALUES (?, ?, ?)
                ON CONFLICT(consumer_id) DO UPDATE SET
                  cursor=excluded.cursor,
                  updated_at=excluded.updated_at;
                """,
                (self._consumer_id, cur, _utc_now_iso()),
            )
